"""
Copyright (c) 2013 Julian Berman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

SPDX-License-Identifier: MIT
"""

# Code is taken from jsonschema package and adapted CloudFormation use
# https://github.com/python-jsonschema/jsonschema
from __future__ import annotations

import logging
from collections import deque
from collections.abc import Mapping
from dataclasses import dataclass, field, fields
from typing import Any, Callable

from cfnlint.conditions import UnknownSatisfisfaction
from cfnlint.context import Context, create_context_for_template
from cfnlint.helpers import is_function
from cfnlint.jsonschema import _keywords, _keywords_cfn, _resolvers_cfn
from cfnlint.jsonschema._filter import FunctionFilter
from cfnlint.jsonschema._format import FormatChecker, cfn_format_checker
from cfnlint.jsonschema._types import TypeChecker, cfn_type_checker
from cfnlint.jsonschema._typing import ResolutionResult, V, ValidationResult
from cfnlint.jsonschema._utils import custom_msg
from cfnlint.jsonschema.exceptions import (
    UndefinedTypeCheck,
    UnknownType,
    ValidationError,
)
from cfnlint.schema.resolver import RefResolver, id_of
from cfnlint.template import Template

LOGGER = logging.getLogger(__name__)


def create(
    validators: Mapping[str, V] | None = None,
    function_filter: FunctionFilter | None = None,
    fn_resolvers: Mapping[str, Any] | None = None,
):
    validators_arg = validators or {}
    function_filter_arg = function_filter or FunctionFilter()
    fn_resolvers_arg = fn_resolvers or {}

    @dataclass
    class Validator:
        """

        Arguments:

            schema:

                The schema that the validator object will validate with.
                It is assumed to be valid, and providing
                an invalid schema can lead to undefined behavior.
        """

        _type_checker: TypeChecker = field(
            init=False, default_factory=lambda: cfn_type_checker
        )
        format_checker: FormatChecker = field(
            init=False, default_factory=lambda: cfn_format_checker
        )

        #: The schema that will be used to validate instances
        schema: Mapping | bool = field(repr=True, default=True)
        validators: Mapping[str, V] = field(
            init=False, default_factory=lambda: validators_arg
        )
        resolver: RefResolver = field(default=None, repr=False)  # type: ignore
        function_filter: FunctionFilter = field(
            init=True, default_factory=lambda: function_filter_arg
        )
        cfn: Template = field(default_factory=lambda: Template(None, {}))
        context: Context = field(default=None)  # type: ignore

        fn_resolvers: Mapping[
            str,
            Callable[["Validator", Any], ResolutionResult],
        ] = field(init=False, default_factory=lambda: fn_resolvers_arg)

        def __post_init__(self):
            if self.context is None:
                self.context = create_context_for_template(self.cfn)
            if self.resolver is None:
                self.resolver = RefResolver.from_schema(
                    schema=self.schema,
                )

        def is_type(self, instance: Any, type: str) -> bool:
            """
            Check if the instance is of the given (JSON Schema) type.

            Arguments:

                instance:

                    the value to check

                type:

                    the name of a known (JSON Schema) type

            Returns:

                whether the instance is of the given type

            Raises:

                `cfnlint.jsonschema.exceptions.UnknownType`:

                    if ``type`` is not a known type
            """
            try:
                return self._type_checker.is_type(instance, type)
            except UndefinedTypeCheck:
                raise UnknownType(type, instance, self.schema)

        def is_valid(self, instance: Any) -> bool:
            """
            Check if the instance is valid under the current `schema`.

            Returns:

                whether the instance is valid or not

            >>> schema = {"maxItems" : 2}
            >>> StandardValidator(schema).is_valid([2, 3, 4])
            False
            """
            error = next(self.iter_errors(instance), None)
            return error is None

        def _resolve_fn(self, key: str, value: Any) -> ResolutionResult:
            for (
                value_resolved_value,
                value_resolved_validator,
                value_resolved_errs,
            ) in self.resolve_value(value):
                if value_resolved_errs:
                    continue
                for (
                    fn_resolved_value,
                    fn_resolved_validator,
                    fn_resolved_err,
                ) in self.fn_resolvers[key](
                    value_resolved_validator, value_resolved_value  # type: ignore
                ):
                    if fn_resolved_err:
                        fn_resolved_err.path.appendleft(key)
                    yield (
                        fn_resolved_value,
                        fn_resolved_validator,
                        fn_resolved_err,
                    )

        def resolve_value(self, instance: Any) -> ResolutionResult:
            key, value = is_function(instance)
            if key in self.fn_resolvers:
                # There is no None in self.fn_resolvers
                for r_value, r_validator, r_errs in self._resolve_fn(key, value):  # type: ignore
                    if not r_errs:
                        try:
                            for _, region_context in r_validator.context.ref_value(
                                "AWS::Region"
                            ):
                                if self.cfn.conditions.satisfiable(
                                    region_context.conditions.status,
                                    region_context.ref_values,
                                ):
                                    yield r_value, r_validator.evolve(
                                        context=region_context.evolve(
                                            is_resolved_value=True,
                                        )
                                    ), r_errs
                        except UnknownSatisfisfaction as err:
                            LOGGER.debug(err)
                            return
                    else:
                        yield None, r_validator, r_errs  # type: ignore
                return

            # The return type is a Protocol and we are returning an instance
            # we will ignore this error
            yield instance, self, None  # type: ignore[misc]

        def iter_errors(self, instance: Any) -> ValidationResult:
            r"""
            Lazily yield each of the validation errors in the given instance.

            >>> schema = {
            ...     "type" : "array",
            ...     "items" : {"enum" : [1, 2, 3]},
            ...     "maxItems" : 2,
            ... }
            >>> v = StandardValidator(schema)
            >>> for error in sorted(v.iter_errors([2, 3, 4]), key=str):
            ...     print(error.message)
            4 is not one of [1, 2, 3]
            [2, 3, 4] is too long
            """
            schema = self.schema

            if schema is True:
                return
            elif schema is False:
                yield ValidationError(
                    f"False schema does not allow {instance!r}",
                    validator=None,
                    validator_value=None,
                    instance=instance,
                    schema=schema,
                )
                return

            scope = id_of(schema)
            if scope:
                self.resolver.push_scope(scope)
            try:
                # we need filter and apply schemas against the new instances
                for _instance, _schema, _validator in self.function_filter.filter(
                    self, instance, schema
                ):
                    for k, v in _schema.items():
                        validator = self.validators.get(k)
                        if validator is None:
                            continue

                        try:
                            for err in (
                                validator(_validator, v, _instance, _schema) or ()
                            ):
                                msg = custom_msg(k, _schema) or err.message
                                if msg is not None:
                                    err.message = msg
                                # set details if not already set by the called fn
                                err._set(
                                    validator=k,
                                    validator_value=v,
                                    instance=_instance,
                                    schema=_schema,
                                    type_checker=self._type_checker,
                                )
                                if k not in {"if", "$ref"}:
                                    err.schema_path.appendleft(k)
                                yield err
                        except Exception as err:
                            LOGGER.debug(err, exc_info=True)
                            yield ValidationError(
                                f"Exception {str(err)!r} raised while validating {k!r}",
                                validator=k,
                                validator_value=v,
                                instance=instance,
                                schema=self.schema,
                                schema_path=deque([k]),
                            )
            finally:
                if scope:
                    self.resolver.pop_scope()

        def validate(self, instance: Any) -> None:
            """
            Check if the instance is valid under the current `schema`.

            Raises:

                `cfnlint.jsonschema.exceptions.ValidationError`:

                    if the instance is invalid

            >>> schema = {"maxItems" : 2}
            >>> StandardValidator(schema).validate([2, 3, 4])
            Traceback (most recent call last):
                ...
            ValidationError: [2, 3, 4] is too long
            """
            for error in self.iter_errors(instance):
                raise error

        def descend(
            self,
            instance: Any,
            schema: Any,
            path: str | int | None = None,
            schema_path: str | int | None = None,
            property_path: str | None = None,
        ) -> ValidationResult:
            for error in self.evolve(
                schema=schema,
                context=self.context.evolve(
                    path=self.context.path.descend(
                        path=path,
                        cfn_path=property_path,
                    ),
                ),
            ).iter_errors(instance):
                if path is not None:
                    error.path.appendleft(path)
                if schema_path is not None:
                    error.schema_path.appendleft(schema_path)
                yield error

        def evolve(self, **kwargs) -> "Validator":
            """
            Create a new validator like this one, but with given changes.

            Preserves all other attributes, so can be used to e.g. create a
            validator with a different schema but with the same :kw:`$ref`
            resolution behavior.

            >>> validator = StandardValidator({})
            >>> validator.evolve(schema={"type": "number"})
            StandardValidator(schema={'type': 'number'}, format_checker=None)
            """
            cls = self.__class__
            for f in fields(Validator):
                if f.init:
                    kwargs.setdefault(f.name, getattr(self, f.name))

            return cls(**kwargs)

        def extend(
            self,
            validators: dict[str, V] | None = None,
            function_filter: FunctionFilter | None = None,
            fn_resolvers: Mapping[str, Callable[["Validator", Any], Any]] | None = None,
        ) -> "Validator":
            """
            Extends the current validator.

            Updates validator in the current instance with the validators provided.

            >>> validator = StandardValidator({}).extend(validators={"type": type})
            """
            all_validators = dict(self.validators)
            if validators is not None:
                all_validators.update(validators)

            if function_filter is None:
                function_filter = self.function_filter

            all_fn_resolvers = dict(self.fn_resolvers)
            if fn_resolvers is not None:
                all_fn_resolvers.update(fn_resolvers)

            return create(  # type: ignore
                validators=all_validators,
                function_filter=function_filter,
                fn_resolvers=all_fn_resolvers,
            )

    return Validator


_standard_validators: dict[str, V] = {
    "$ref": _keywords.ref,
    "additionalProperties": _keywords.additionalProperties,
    "allOf": _keywords.allOf,
    "anyOf": _keywords.anyOf,
    "const": _keywords.const,
    "contains": _keywords.contains,
    "dependencies": _keywords.dependencies,
    "dependentRequired": _keywords.dependentRequired,
    "dependentExcluded": _keywords.dependentExcluded,
    "enum": _keywords.enum,
    "exclusiveMaximum": _keywords.exclusiveMaximum,
    "exclusiveMinimum": _keywords.exclusiveMinimum,
    "format": _keywords.format,
    "if": _keywords.if_,
    "items": _keywords.items,
    "maxItems": _keywords.maxItems,
    "maxLength": _keywords.maxLength,
    "maxProperties": _keywords.maxProperties,
    "maximum": _keywords.maximum,
    "minItems": _keywords.minItems,
    "minLength": _keywords.minLength,
    "minProperties": _keywords.minProperties,
    "minimum": _keywords.minimum,
    "multipleOf": _keywords.multipleOf,
    "not": _keywords.not_,
    "oneOf": _keywords.oneOf,
    "pattern": _keywords.pattern,
    "patternProperties": _keywords.patternProperties,
    "prefixItems": _keywords.prefixItems,
    "properties": _keywords.properties,
    "propertyNames": _keywords.propertyNames,
    "required": _keywords.required,
    "requiredXor": _keywords.requiredXor,
    "type": _keywords.type,
    "uniqueItems": _keywords.uniqueItems,
    "uniqueKeys": _keywords.uniqueKeys,
}

CfnTemplateValidator = create(
    validators={
        **_standard_validators,
        **_keywords_cfn.cfn_validators,
    },
    function_filter=FunctionFilter(),
    fn_resolvers=_resolvers_cfn.fn_resolvers,
)

StandardValidator = create(
    validators=_standard_validators,
    function_filter=FunctionFilter(),
)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.functions.DynamicReferenceSecureString import (
    DynamicReferenceSecureString,
)


@pytest.fixture(scope="module")
def rule():
    rule = DynamicReferenceSecureString()
    yield rule


@pytest.fixture
def template():
    return {
        "Resources": {
            "MyResource": {
                "Type": "AWS::IAM::User",
                "Properties": {"LoginProfile": {"Password": "Foo"}},
            }
        }
    }


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "Valid SSM Secure Parameter",
            "{{resolve:ssm-secure:Parameter}}",
            {
                "cfn_path": [
                    "Resources",
                    "AWS::IAM::User",
                    "Properties",
                    "LoginProfile",
                    "Password",
                ]
            },
            [],
        ),
        (
            "Invalid SSM secure location",
            "{{resolve:ssm-secure:Parameter}}",
            {"cfn_path": ["Outputs", "*", "Value"]},
            [
                ValidationError(
                    (
                        "Dynamic reference '{{resolve:ssm-secure:Parameter}}' "
                        "to SSM secure strings can only be used in resource properties"
                    ),
                    rule=DynamicReferenceSecureString(),
                )
            ],
        ),
    ],
    indirect=["path"],
)
def test_validate(name, instance, path, expected, rule, validator):

    errs = list(rule.validate(validator, {"type": "string"}, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
import sys
from pathlib import Path
from test.testlib.testcase import BaseTestCase

import defusedxml.ElementTree as ET

import cfnlint.config
import cfnlint.formatters
import cfnlint.helpers
from cfnlint import ConfigMixIn
from cfnlint.formatters import (
    Formatter,
    JsonFormatter,
    JUnitFormatter,
    ParseableFormatter,
    PrettyFormatter,
    QuietFormatter,
    SARIFFormatter,
)
from cfnlint.jsonschema import StandardValidator
from cfnlint.rules import Match, Rules
from cfnlint.rules.functions.SubUnneeded import SubUnneeded
from cfnlint.rules.resources.properties.Type import Type
from cfnlint.rules.resources.UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes import (
    UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes,
)


class TestFormatters(BaseTestCase):
    """Test Formatters"""

    def setUp(self) -> None:
        super().setUp()
        self.rules = Rules.create_from_directory(cfnlint.config._DEFAULT_RULESDIR)
        self.filename = str(Path("test/fixtures/templates/bad/formatters.yaml"))
        self.config = ConfigMixIn(
            cli_args=[
                "--include-checks",
                "I",
                "--ignore-checks",
                "E1029",
                "--",
                self.filename,
            ]
        )

        self.results = [
            Match(
                linenumber=6,
                columnnumber=3,
                linenumberend=6,
                columnnumberend=10,
                filename=self.filename,
                rule=UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes(),
                message=(
                    "The default action when replacing/removing a resource is to delete"
                    " it. Set explicit values for UpdateReplacePolicy / DeletionPolicy"
                    " on potentially stateful resource: Resources/myTable"
                ),
            ),
            Match(
                linenumber=9,
                columnnumber=7,
                linenumberend=9,
                columnnumberend=16,
                filename=self.filename,
                rule=SubUnneeded(),
                message=(
                    "Fn::Sub isn't needed because there are no variables at"
                    " Resources/myTable/Properties/TableName/Fn::Sub"
                ),
            ),
            Match(
                linenumber=18,
                columnnumber=9,
                linenumberend=18,
                columnnumberend=27,
                filename=self.filename,
                rule=Type(),
                message="'5' is not of type 'integer'",
            ),
        ]

    def test_base_formatter(self):
        """Test base formatter"""
        formatter = Formatter()
        a_results = formatter.print_matches(
            self.results, self.rules, self.config
        ).splitlines()

        # Check the errors
        self.assertEqual(
            a_results[0], f"{self.results[0].rule.id} {self.results[0].message}"
        )
        self.assertEqual(
            a_results[1],
            f"{self.results[0].filename}:{self.results[0].linenumber}:{self.results[0].columnnumber}",
        )
        self.assertEqual(a_results[2], "")
        self.assertEqual(
            a_results[3], f"{self.results[1].rule.id} {self.results[1].message}"
        )
        self.assertEqual(
            a_results[4],
            f"{self.results[1].filename}:{self.results[1].linenumber}:{self.results[1].columnnumber}",
        )
        self.assertEqual(a_results[5], "")
        self.assertEqual(
            a_results[6], f"{self.results[2].rule.id} {self.results[2].message}"
        )
        self.assertEqual(
            a_results[7],
            f"{self.results[2].filename}:{self.results[2].linenumber}:{self.results[2].columnnumber}",
        )

    def test_quiet_formatter(self):
        """Test quiet formatter"""

        formatter = QuietFormatter()
        a_results = formatter.print_matches(
            self.results, self.rules, self.config
        ).splitlines()

        for i in range(3):
            # Check the errors
            # ruff: noqa: E501
            self.assertEqual(
                a_results[i],
                (
                    f"{self.results[i].rule.id}:"
                    f" {self.results[i].rule.shortdesc} {self.results[i].filename}:{self.results[i].linenumber}"
                ),
            )

    def test_parseable_formatter(self):
        """Test Parseable formatter"""
        formatter = ParseableFormatter()
        results = formatter.print_matches(
            self.results, self.rules, self.config
        ).splitlines()

        for i in range(3):
            # Check the errors
            self.assertEqual(
                results[i],
                f"{self.results[i].filename}:{self.results[i].linenumber}:{self.results[i].columnnumber}:{self.results[i].linenumberend}:{self.results[i].columnnumberend}:{self.results[i].rule.id}:{self.results[i].message}",
            )

    def test_pretty_formatter(self):
        """Test pretty formatter"""
        formatter = PrettyFormatter()
        results = formatter.print_matches(
            self.results, rules=self.rules, config=self.config
        ).splitlines()

        if sys.stdout.isatty():
            # Check the errors
            self.assertEqual(results[0], f"\x1b[4m{self.filename}\x1b[0m")
            self.assertEqual(
                results[1],
                (
                    f"\x1b[0m{self.results[2].linenumber}:{self.results[2].columnnumber}:"
                    f"               \x1b[0m\x1b[31m{self.results[2].rule.id}    "
                    f" \x1b[0m{self.results[2].message}"
                ),
            )
            self.assertEqual(
                results[2],
                (
                    f"\x1b[0m{self.results[1].linenumber}:{self.results[1].columnnumber}:"
                    f"                \x1b[0m\x1b[33m{self.results[1].rule.id}    "
                    f" \x1b[0m{self.results[1].message}"
                ),
            )
            self.assertEqual(
                results[3],
                (
                    f"\x1b[0m{self.results[0].linenumber}:{self.results[0].columnnumber}:"
                    f"                \x1b[0m\x1b[34m{self.results[0].rule.id}    "
                    f" \x1b[0m{self.results[0].message}"
                ),
            )
        else:
            # Check the errors
            self.assertEqual(results[0], self.filename)
            self.assertEqual(
                results[1],
                (
                    f"{self.results[2].linenumber}:{self.results[2].columnnumber}:     "
                    f"          {self.results[2].rule.id}     {self.results[2].message}"
                ),
            )
            self.assertEqual(
                results[2],
                (
                    f"{self.results[1].linenumber}:{self.results[1].columnnumber}:     "
                    f"           {self.results[1].rule.id}    "
                    f" {self.results[1].message}"
                ),
            )
            self.assertEqual(
                results[3],
                (
                    f"{self.results[0].linenumber}:{self.results[0].columnnumber}:     "
                    f"           {self.results[0].rule.id}    "
                    f" {self.results[0].message}"
                ),
            )

    def test_pretty_formatter_pipe(self):
        """Test pretty formatter"""
        formatter = PrettyFormatter()
        self.config.cli_args.templates = None
        results = formatter.print_matches(
            self.results, rules=self.rules, config=self.config
        ).splitlines()

        if sys.stdout.isatty():
            self.assertIn("Cfn-lint scanned 1 templates", results[5])
        else:
            # Check the errors
            self.assertIn("Cfn-lint scanned 1 templates", results[5])

    def test_json_formatter(self):
        """Test JSON formatter"""
        formatter = JsonFormatter()

        # Get the JSON output
        json_results = json.loads(
            formatter.print_matches(self.results, self.rules, self.config)
        )

        # Check the 3 errors again
        self.assertEqual(len(json_results), 3)

        # Check the errors level description
        for i in range(3):
            self.assertEqual(json_results[i]["Rule"]["Id"], self.results[i].rule.id)
            self.assertEqual(json_results[i]["Message"], self.results[i].message)

    def test_junit_returns_none(self):
        """Test JUnut Formatter returns None if no rules are passed in"""
        formatter = JUnitFormatter()

        # The actual test
        self.assertIsNone(formatter.print_matches([], [], self.config))

    def test_junit_formatter(self):
        """Test JUnit Formatter"""
        formatter = JUnitFormatter()

        self.rules._used_rules = {
            "I3011": self.rules["I3011"],
            "E3012": self.rules["E3012"],
            "W1020": self.rules["W1020"],
        }
        s = formatter.print_matches(self.results, self.rules, self.config)
        root = ET.fromstring(s)

        self.assertEqual(root.tag, "testsuites")
        self.assertEqual(root[0].tag, "testsuite")

        found_i3011 = False
        found_w1020 = False
        found_e3012 = False
        name_i3011 = "{0} {1}".format(
            self.results[0].rule.id, self.results[0].rule.shortdesc
        )
        name_w1020 = "{0} {1}".format(
            self.results[1].rule.id, self.results[1].rule.shortdesc
        )
        name_e3012 = "{0} {1}".format(
            self.results[2].rule.id, self.results[2].rule.shortdesc
        )
        name_e1029 = "E1029 Sub is required if a variable is used in a string"
        for child in root[0]:
            self.assertEqual(child.tag, "testcase")

            if child.attrib["name"] in [name_i3011, name_w1020, name_e3012]:
                self.assertEqual(child[0].tag, "failure")

                if child.attrib["name"] == name_i3011:
                    found_i3011 = True
                if child.attrib["name"] == name_w1020:
                    found_w1020 = True
                if child.attrib["name"] == name_e3012:
                    found_e3012 = True

            if child.attrib["name"] == name_e1029:
                self.assertEqual(child[0].tag, "skipped")
                self.assertEqual(child[0].attrib["type"], "skipped")

        self.assertTrue(found_i3011)
        self.assertTrue(found_w1020)
        self.assertTrue(found_e3012)

    def test_sarif_formatter(self):
        """Test the SARIF formatter"""
        formatter = SARIFFormatter()

        # Get the SARIF JSON output
        sarif = json.loads(
            formatter.print_matches(self.results, self.rules, self.config)
        )

        # Fetch the SARIF schema
        schema = json.loads(cfnlint.helpers.get_url_content(sarif["$schema"], False))
        validator = StandardValidator(schema=schema)
        validator.validate(sarif)

        sarif_results = sarif["runs"][0]["results"]
        # Check the 3 errors again
        self.assertEqual(len(sarif_results), 3)

        # Sanity check the errors
        self.assertEqual(sarif_results[0]["level"], "note")
        self.assertEqual(sarif_results[0]["ruleId"], "I3011")
        # IMPORTANT: 'warning' is the default level in SARIF (when kind is
        # absent) and is stripped by serialization
        self.assertNotIn("level", sarif_results[1].keys())
        self.assertEqual(sarif_results[1]["ruleId"], "W1020")
        self.assertEqual(sarif_results[2]["level"], "error")
        self.assertEqual(sarif_results[2]["ruleId"], "E3012")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.rds.DbClusterEngineVersion import DbClusterEngineVersion


@pytest.fixture(scope="module")
def rule():
    rule = DbClusterEngineVersion()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "Engine": "mysql",
            },
            [],
        ),
        (
            {
                "Engine": "mysql",
                "EngineVersion": "5.7.44",
            },
            [],
        ),
        (
            {
                "Engine": {"Ref": "Engine"},
                "EngineVersion": {"Ref": "EngineVersion"},
            },
            [],
        ),
        (
            {
                "Engine": "mysql",
                "EngineVersion": {"Ref": "EngineVersion"},
            },
            [],
        ),
        (
            {"Engine": "foo"},
            [
                ValidationError(
                    (
                        "'foo' is not one of ['aurora-mysql', 'aurora-postgresql', "
                        "'mysql', 'postgres']"
                    ),
                    rule=DbClusterEngineVersion(),
                    path=deque(["Engine"]),
                    validator="enum",
                    schema_path=deque(
                        ["allOf", 0, "then", "properties", "Engine", "enum"]
                    ),
                )
            ],
        ),
        (
            {"Engine": "mysql", "EngineVersion": "foo"},
            [
                ValidationError(
                    (
                        "'foo' is not one of ['5.7.44', '5.7.44-rds.20240408', "
                        "'5.7.44-rds.20240529', '8.0.32', '8.0.33', '8.0.34', "
                        "'8.0.35', '8.0.36', '8.0.37', '8.0.39']"
                    ),
                    rule=DbClusterEngineVersion(),
                    path=deque(["EngineVersion"]),
                    validator="enum",
                    schema_path=deque(
                        ["allOf", 3, "then", "properties", "EngineVersion", "enum"]
                    ),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque

import pytest

from cfnlint.rules.helpers import get_value_from_path


@pytest.fixture
def template():
    return {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Parameters": {
            "ImageId": {
                "Type": "AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>",
                "Default": "/aws/service/ami-amazon-linux-latest/amzn-ami-hvm-x86_64-gp2",  # noqa: E501
            }
        },
        "Conditions": {
            "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
            "IsNotUsEast1": {"Fn::Not": [{"Condition": "IsUsEast1"}]},
            "IsImageIdSpecified": {
                "Fn::Not": [{"Fn::Equals": [{"Ref": "ImageId"}, ""]}]
            },
        },
        "Resources": {},
    }


@pytest.mark.parametrize(
    "name,instance,v_path,status,expected",
    [
        (
            "No path",
            {"Foo": "Bar"},
            deque([]),
            {},
            [
                ({"Foo": "Bar"}, {}, deque([])),
            ],
        ),
        (
            "Small path with out it being found",
            {"Foo": "Bar"},
            deque(["Bar"]),
            {},
            [
                (None, {}, deque(["Bar"])),
            ],
        ),
        (
            "Bad path returns an empty value with list",
            {"Foo": [{"Bar": "One"}]},
            deque(["Foo", "Bar"]),
            {},
            [
                (
                    None,
                    {},
                    deque(["Foo"]),
                )
            ],
        ),
        (
            "Bad path returns an empty value with string",
            {"Foo": "Misvalued"},
            deque(["Foo", "Bar"]),
            {},
            [
                (
                    None,
                    {},
                    deque(["Foo"]),
                )
            ],
        ),
        (
            "With a path and no conditions",
            {"Foo": {"Bar": True}},
            deque(["Foo", "Bar"]),
            {},
            [
                (True, {}, deque(["Foo", "Bar"])),
            ],
        ),
        (
            "With a list item",
            {"Foo": [{"Bar": "One"}, {"Bar": "Two"}, {"NoValue": "Three"}]},
            deque(["Foo", "*", "Bar"]),
            {},
            [
                ("One", {}, deque(["Foo", 0, "Bar"])),
                ("Two", {}, deque(["Foo", 1, "Bar"])),
                (None, {}, deque(["Foo", 2, "Bar"])),
            ],
        ),
        (
            "With a basic condition",
            {
                "Foo": {
                    "Fn::If": [
                        "IsUsEast1",
                        {"Bar": "One"},
                        {"Bar": "Two"},
                    ]
                }
            },
            deque(["Foo", "Bar"]),
            {},
            [
                ("One", {"IsUsEast1": True}, deque(["Foo", "Fn::If", 1, "Bar"])),
                ("Two", {"IsUsEast1": False}, deque(["Foo", "Fn::If", 2, "Bar"])),
            ],
        ),
        (
            "With a basic condition and a current status",
            {
                "Foo": {
                    "Fn::If": [
                        "IsUsEast1",
                        {"Bar": "One"},
                        {"Bar": "Two"},
                    ]
                }
            },
            deque(["Foo", "Bar"]),
            {"IsUsEast1": True},
            [
                ("One", {"IsUsEast1": True}, deque(["Foo", "Fn::If", 1, "Bar"])),
            ],
        ),
        (
            "With a basic condition and a current status on a related condition",
            {
                "Foo": {
                    "Fn::If": [
                        "IsUsEast1",
                        {"Bar": "One"},
                        {"Bar": "Two"},
                    ]
                }
            },
            deque(["Foo", "Bar"]),
            {"IsNotUsEast1": True},
            [
                (
                    "Two",
                    {"IsUsEast1": False, "IsNotUsEast1": True},
                    deque(["Foo", "Fn::If", 2, "Bar"]),
                ),
            ],
        ),
        (
            "With a random function in the way",
            {"Foo": {"Fn::FindInMap": ["A", "B", "C"]}},
            deque(["Foo", "*", "Bar"]),
            {},
            [],
        ),
        (
            "With a ref at the desination",
            {"Foo": {"Ref": "Bar"}},
            deque(["Foo"]),
            {},
            [({"Ref": "Bar"}, {}, deque(["Foo"]))],
        ),
    ],
)
def test_get_value_from_path(name, instance, v_path, status, expected, validator):
    validator = validator.evolve(
        context=validator.context.evolve(
            conditions=validator.context.conditions.evolve(
                status=status,
            ),
        ),
    )

    results = list(get_value_from_path(validator, instance, v_path))

    assert len(results) == len(expected), f"Test {name!r} got {len(results)!r}"
    for result, exp in zip(results, expected):
        assert result[0] == exp[0], f"Test {name!r} got {result[0]!r}"
        assert (
            result[1].context.conditions.status == exp[1]
        ), f"Test {name!r} got {result[1].context.conditions.status!r}"
        assert (
            result[1].context.path.path == exp[2]
        ), f"Test {name!r} got {result[1].context.path.path!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Context, Path
from cfnlint.context.context import Parameter, Resource
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.resources.DeletionPolicy import DeletionPolicy


@pytest.mark.parametrize(
    "name, instance, path, expected",
    [
        ("Correct type", "Snapshot", deque(["Resources", "MyVolume"]), []),
        (
            "Incorrect type",
            {"Foo": "Bar"},
            deque(["Resources", "MyVolume"]),
            [
                ValidationError(
                    "{'Foo': 'Bar'} is not of type 'string'",
                    schema_path=deque(["type"]),
                    validator="type",
                    rule=DeletionPolicy(),
                ),
                ValidationError(
                    (
                        "{'Foo': 'Bar'} is not one of ['Delete', 'Retain', "
                        "'RetainExceptOnCreate', 'Snapshot']"
                    ),
                    schema_path=deque(["enum"]),
                    validator="enum",
                    rule=DeletionPolicy(),
                ),
            ],
        ),
        (
            "Snapshot not supported",
            "Snapshot",
            deque(["Resources", "MyInstance"]),
            [
                ValidationError(
                    (
                        "'Snapshot' is not one of ['Delete', 'Retain', "
                        "'RetainExceptOnCreate']"
                    ),
                    schema_path=deque(["enum"]),
                    validator="enum",
                    rule=DeletionPolicy(),
                ),
            ],
        ),
        (
            "Success on a valid function",
            {"Ref": "MyParameter"},
            deque(["Resources", "MyInstance"]),
            [],
        ),
    ],
)
def test_deletion_policy(name, instance, path, expected):
    context = Context(
        regions=["us-east-1"],
        path=Path(path=path),
        resources={
            "MyVolume": Resource({"Type": "AWS::EC2::Volume"}),
            "MyInstance": Resource({"Type": "AWS::EC2::Instance"}),
        },
        parameters={"MyParameter": Parameter({"Type": "String"})},
    )
    validator = CfnTemplateValidator(context=context)

    rule = DeletionPolicy()
    errors = list(
        rule.validate(
            validator=validator,
            dP="deletionpolicy",
            instance=instance,
            schema={},
        )
    )

    assert errors == expected



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint._typing import RuleMatches
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.template import Template


class RelationshipConditions(CloudFormationLintRule):
    """Check if Ref/GetAtt values are available via conditions"""

    id = "W1001"
    shortdesc = "Ref/GetAtt to resource that is available when conditions are applied"
    description = (
        "Check the Conditions that affect a Ref/GetAtt to make sure "
        "the resource being related to is available when there is a resource "
        "condition."
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-ref.html"
    tags = ["conditions", "resources", "relationships", "ref", "getatt", "sub"]

    def match(self, cfn: Template) -> RuleMatches:
        """Check CloudFormation Ref/GetAtt for Conditions"""

        matches: RuleMatches = []
        if cfn.graph is None:
            return matches
        for edge in cfn.graph.graph.edges.data():
            source, destination, data = edge

            # Depends on is already covered
            if data.get("label") == "DependsOn":
                continue
            # Double check the destination is a in the graph
            if destination not in cfn.graph.graph.nodes:
                continue
            # If the destination isn't a resource it won't matter
            if "Resource" != cfn.graph.graph.nodes[destination]["type"]:
                continue

            path = [
                f"{cfn.graph.graph.nodes[source].get('type')}s",
                source.replace("Output-", ""),
            ] + data["source_paths"]
            scenarios = cfn.is_resource_available(path, destination)

            for scenario in scenarios:
                # pylint: disable=consider-using-f-string
                scenario_text = " and ".join(
                    [
                        "when condition '%s' is %s" % (k, v)
                        for (k, v) in scenario.items()
                    ]
                )

                message = (
                    f"{data.get('label')} to resource "
                    f"{destination!r} that may not be available "
                    f"{scenario_text} at {'/'.join(map(str, path))}"
                )
                matches.append(RuleMatch(path, message))

        return matches



"""
Copyright (c) 2013 Julian Berman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

SPDX-License-Identifier: MIT
"""

# Code is taken from jsonschema package and adapted CloudFormation use
# https://github.com/python-jsonschema/jsonschema

from __future__ import annotations

from copy import deepcopy
from difflib import SequenceMatcher
from typing import Any, Sequence

import regex as re

from cfnlint.helpers import ensure_list
from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.jsonschema._typing import ValidationResult
from cfnlint.jsonschema._utils import (
    equal,
    find_additional_properties,
    unbool,
    uniq,
    uniq_keys,
)
from cfnlint.jsonschema.exceptions import FormatError


def additionalProperties(
    validator: Validator, aP: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if not validator.is_type(instance, "object"):
        return

    extras = set(find_additional_properties(validator, instance, schema))
    if validator.is_type(aP, "object"):
        for extra in extras:
            yield from validator.descend(
                instance[extra],
                aP,
                path=extra,
                property_path="*",
            )
    elif not aP and extras:
        if "patternProperties" in schema:
            patterns = ", ".join(
                repr(each) for each in sorted(schema["patternProperties"])
            )
            for extra in extras:
                error = f"{extra!r} does not match any of the regexes: {patterns}"
                yield ValidationError(error, path=[extra])
        else:
            for extra in extras:
                for key in schema.get("properties", {}).keys():
                    if SequenceMatcher(a=extra, b=key).ratio() > 0.8:
                        yield ValidationError(
                            (
                                f"Additional properties are not allowed ({extra!r} "
                                f"was unexpected. Did you mean {key!r}?)"
                            ),
                            path=[extra],
                        )
                        break
                else:
                    yield ValidationError(
                        (
                            f"Additional properties are not allowed ({extra!r} "
                            "was unexpected)"
                        ),
                        path=[extra],
                    )


def allOf(
    validator: Validator, allOf: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    validator = validator.evolve(
        function_filter=validator.function_filter.evolve(
            add_cfn_lint_keyword=False,
        )
    )
    for index, subschema in enumerate(allOf):
        yield from validator.descend(instance, subschema, schema_path=index)


def anyOf(
    validator: Validator, anyOf: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    validator = validator.evolve(
        function_filter=validator.function_filter.evolve(
            add_cfn_lint_keyword=False,
        )
    )
    all_errors = []
    other_errors = []
    for index, subschema in enumerate(anyOf):
        errs = []
        # warning and informational shouldn't count towards if anyOf is
        # valid.  Save W, I errors and return if errors exist
        for err in validator.descend(instance, subschema, schema_path=index):
            if err.rule is not None and not err.rule.id.startswith("E"):
                other_errors.append(err)
                continue
            errs.append(err)
        if not errs:
            break
        all_errors.extend(errs)
    else:
        yield ValidationError(
            f"{instance!r} is not valid under any of the given schemas",
            context=all_errors + other_errors,
        )


def const(
    validator: Validator, const: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if not equal(instance, const):
        yield ValidationError(f"{const!r} was expected")


def contains(
    validator: Validator, contains: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if not validator.is_type(instance, "array"):
        return

    if not any(
        validator.evolve(schema=contains).is_valid(element) for element in instance
    ):
        yield ValidationError(
            f"{instance!r} does not contain items matching the given schema",
        )


def dependencies(
    validator: Validator, dependencies: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if not validator.is_type(instance, "object"):
        return

    for property, dependency in dependencies.items():
        if property not in instance:
            continue

        if validator.is_type(dependency, "array"):
            for each in dependency:
                if each not in instance:
                    message = f"{each!r} is a dependency of {property!r}"
                    yield ValidationError(message)
        else:
            yield from validator.descend(
                instance,
                dependency,
                schema_path=property,
            )


def dependentRequired(
    validator: Validator,
    dependentRequired: dict[str, Sequence[str]],
    instance: Any,
    schema: dict[str, Any],
):
    if not validator.is_type(instance, "object"):
        return

    for property, dependency in dependentRequired.items():
        if property not in instance:
            continue

        for each in dependency:
            if each not in instance:
                message = f"{each!r} is a dependency of {property!r}"
                yield ValidationError(message)


def dependentExcluded(
    validator: Validator,
    dependentExcluded: dict[str, Sequence[str]],
    instance: Any,
    schema: dict[str, Any],
):
    if not validator.is_type(instance, "object"):
        return

    for property, dependency in dependentExcluded.items():
        if property not in instance:
            continue

        for each in dependency:
            if each in instance:
                message = f"{each!r} should not be included with {property!r}"
                yield ValidationError(message, path=[each])


def enum(
    validator: Validator, enums: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if instance in (0, 1):
        unbooled = unbool(instance)
        if all(not (equal(unbooled, unbool(each))) for each in enums):
            yield ValidationError(f"{instance!r} is not one of {enums!r}")
    else:
        if all(not (equal(instance, each)) for each in enums):
            yield ValidationError(f"{instance!r} is not one of {enums!r}")


def exclusiveMaximum(
    validator: Validator, m: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    t_instance = deepcopy(instance)
    if validator.is_type(t_instance, "string"):
        try:
            t_instance = float(t_instance)
        except ValueError:
            return
    if not validator.is_type(t_instance, "number"):
        return

    if t_instance >= m:
        yield ValidationError(
            f"{instance!r} is greater than or equal to the maximum of {m!r}",
        )


def exclusiveMinimum(
    validator: Validator, m: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    t_instance = deepcopy(instance)
    if validator.is_type(t_instance, "string"):
        try:
            t_instance = float(t_instance)
        except ValueError:
            return
    if not validator.is_type(t_instance, "number"):
        return

    if t_instance <= m:
        yield ValidationError(
            f"{instance!r} is less than or equal to the minimum of {m!r}",
        )


def format(
    validator: Validator, format: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if validator.format_checker is not None:  # type: ignore
        try:
            validator.format_checker.check(instance, format)  # type: ignore
        except FormatError as error:
            yield ValidationError(error.message, cause=error.cause)


def if_(
    validator: Validator, if_schema: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    validator = validator.evolve(
        function_filter=validator.function_filter.evolve(
            add_cfn_lint_keyword=False,
        )
    )
    if validator.evolve(schema=if_schema).is_valid(instance):
        if "then" in schema:
            then = schema["then"]
            yield from validator.descend(instance, then, schema_path="then")
    elif "else" in schema:
        else_ = schema["else"]
        yield from validator.descend(instance, else_, schema_path="else")


def items(
    validator: Validator, items: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if not validator.is_type(instance, "array"):
        return

    if validator.is_type(items, "array"):
        for (index, item), subschema in zip(enumerate(instance), items):
            yield from validator.descend(
                item,
                subschema,
                path=index,
                schema_path=index,
                property_path="*",
            )
    else:
        for index, item in enumerate(instance):
            yield from validator.descend(item, items, path=index, property_path="*")


def maxItems(
    validator: Validator, mI: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:  # pylint: disable=arguments-renamed
    if validator.is_type(instance, "array") and len(instance) > mI:
        yield ValidationError(f"{instance!r} is too long ({mI})")


def maxLength(
    validator: Validator, mL: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:  # pylint: disable=arguments-renamed
    if not validator.is_type(instance, "string"):
        return

    if len(instance) > mL:
        yield ValidationError(f"{instance!r} is longer than {mL}")


def maxProperties(
    validator: Validator, mP: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if not validator.is_type(instance, "object"):
        return
    if validator.is_type(instance, "object") and len(instance) > mP:
        yield ValidationError(f"{instance!r} has too many properties")


def maximum(
    validator: Validator, m: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    t_instance = deepcopy(instance)
    if validator.is_type(t_instance, "string"):
        try:
            t_instance = float(t_instance)
        except:  # noqa: E722
            return
    if not validator.is_type(t_instance, "number"):
        return

    if t_instance > m:
        message = f"{instance!r} is greater than the maximum of {m!r}"
        yield ValidationError(message)


def minItems(
    validator: Validator, mI: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if validator.is_type(instance, "array") and len(instance) < mI:
        yield ValidationError(f"{instance!r} is too short ({mI})")


def minLength(
    validator: Validator, mL: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:  # pylint: disable=arguments-renamed
    if not validator.is_type(instance, "string"):
        return

    if len(instance) < mL:
        yield ValidationError(f"{instance!r} is shorter than {mL}")


def minProperties(
    validator: Validator, mP: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if validator.is_type(instance, "object") and len(instance) < mP:
        yield ValidationError(f"{instance!r} does not have enough properties")


def minimum(
    validator: Validator, m: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    t_instance = deepcopy(instance)
    if validator.is_type(t_instance, "string"):
        try:
            t_instance = float(t_instance)
        except ValueError:
            return
    if not validator.is_type(t_instance, "number"):
        return

    if t_instance < m:
        message = f"{instance!r} is less than the minimum of {m!r}"
        yield ValidationError(message)


def multipleOf(
    validator: Validator, dB: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    t_instance = deepcopy(instance)
    if validator.is_type(t_instance, "string"):
        try:
            t_instance = float(t_instance)
        except ValueError:
            return
    if not validator.is_type(t_instance, "number"):
        return

    if isinstance(dB, float):
        quotient = t_instance / dB
        try:
            failed = int(quotient) != quotient
        except OverflowError:
            pass
    else:
        failed = t_instance % dB

    if failed:
        yield ValidationError(f"{instance!r} is not a multiple of {dB}")


def not_(
    validator: Validator, not_schema: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    validator = validator.evolve(
        function_filter=validator.function_filter.evolve(
            add_cfn_lint_keyword=False,
        )
    )
    if validator.evolve(schema=not_schema).is_valid(instance):
        message = f"{instance!r} should not be valid under {not_schema!r}"
        yield ValidationError(message)


def oneOf(
    validator: Validator, oneOf: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    validator = validator.evolve(
        function_filter=validator.function_filter.evolve(
            add_cfn_lint_keyword=False,
        )
    )
    subschemas = enumerate(oneOf)
    all_errors = []
    for index, subschema in subschemas:
        errs = list(validator.descend(instance, subschema, schema_path=index))
        if not errs:
            first_valid = subschema
            break
        all_errors.extend(errs)
    else:
        yield ValidationError(
            f"{instance!r} is not valid under any of the given schemas",
            context=all_errors,
        )

    more_valid = [
        each
        for _, each in subschemas
        if validator.evolve(schema=each).is_valid(instance)
    ]
    if more_valid:
        more_valid.append(first_valid)
        reprs = ", ".join(repr(schema) for schema in more_valid)
        yield ValidationError(f"{instance!r} is valid under each of {reprs}")


def pattern(
    validator: Validator, patrn: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if not validator.is_type(instance, "string"):
        return

    if not re.search(patrn, instance):
        yield ValidationError(f"{instance!r} does not match {patrn!r}")


def patternProperties(
    validator: Validator, patternProperties: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if not validator.is_type(instance, "object"):
        return

    for pattern, subschema in patternProperties.items():
        for k, v in instance.items():
            if not validator.is_type(k, "string"):
                continue
            if re.search(pattern, k):
                yield from validator.descend(
                    v,
                    subschema,
                    path=k,
                    schema_path=pattern,
                    property_path="*",
                )


def properties(
    validator: Validator, properties: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if not validator.is_type(instance, "object"):
        return

    for p, subschema in properties.items():
        # use the instance keys because it gives us the start_mark
        k = [k for k in instance.keys() if k == p]
        if p in instance:
            yield from validator.descend(
                instance[p],
                subschema,
                path=k[0] if k else p,
                schema_path=p,
                property_path=p,
            )


def propertyNames(
    validator: Validator, propertyNames: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if not validator.is_type(instance, "object"):
        return

    for property in instance:
        yield from validator.descend(
            instance=property,
            schema=propertyNames,
            path=property,
        )


def ref(
    validator: Validator, ref: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    scope, resolved = validator.resolver.resolve(ref)
    validator.resolver.push_scope(scope)

    try:
        yield from validator.descend(instance, resolved)
    finally:
        validator.resolver.pop_scope()


def required(
    validator: Validator, required: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if not validator.is_type(instance, "object"):
        return
    for property in required:
        if property not in instance:
            yield ValidationError(f"{property!r} is a required property")


def requiredXor(
    validator: Validator, required: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if not validator.is_type(instance, "object"):
        return
    matches = set(required).intersection(instance.keys())
    if not matches:
        yield ValidationError(f"Only one of {required!r} is a required property")
        return
    if len(matches) > 1:
        for match in matches:
            yield ValidationError(
                f"Only one of {required!r} is a required property", path=[match]
            )


def uniqueItems(
    validator: Validator, uI: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if uI and validator.is_type(instance, "array") and not uniq(instance):
        yield ValidationError(f"{instance!r} has non-unique elements")


def uniqueKeys(
    validator: Validator, uKs: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    if uKs and validator.is_type(instance, "array") and not uniq_keys(instance, uKs):
        yield ValidationError(f"{instance!r} has non-unique elements for keys {uKs!r}")


def type(
    validator: Validator, tS: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    tS = ensure_list(tS)
    if not any(validator.is_type(instance, type) for type in tS):
        reprs = ", ".join(repr(type) for type in tS)
        yield ValidationError(f"{instance!r} is not of type {reprs}")


def prefixItems(validator: Validator, prefixItems: Any, instance: Any, schema: Any):
    if not validator.is_type(instance, "array"):
        return

    for (index, item), subschema in zip(enumerate(instance), prefixItems):
        yield from validator.descend(
            instance=item,
            schema=subschema,
            schema_path=index,
            path=index,
            property_path="*",
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from test.unit.rules import BaseRuleTestCase

from cfnlint.context import create_context_for_template
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.rules.jsonschema.Base import BaseJsonSchema
from cfnlint.template import Template


class RuleWithFunction(CloudFormationLintRule):
    """Test Rule"""

    id = "E3XXX"
    shortdesc = "Test Rule"
    description = "Test Rule"
    source_url = ""
    tags: list[str] = []

    def bar(validator, bar, instance, schema):
        pass


class RuleWithOutFunction(CloudFormationLintRule):
    """Test Rule"""

    id = "E3YYY"
    shortdesc = "Test Rule"
    description = "Test Rule"
    source_url = ""
    tags: list[str] = []


class TestBaseJsonSchema(BaseRuleTestCase):
    """Test Base Json Schema"""

    def test_extend_validator(self):
        def _validator(validator, t, instance, schema):
            pass

        class Rule(BaseJsonSchema):
            def __init__(self) -> None:
                super().__init__()
                self.rule_set = {
                    "bar": "EXXXX",
                    "foobar": "EYYYY",
                }
                self.child_rules = {
                    "EXXXX": RuleWithFunction(),
                    "EYYYY": RuleWithOutFunction(),
                }
                self.validators = {
                    "foo": _validator,
                    "bar": None,
                    "foobar": None,
                }

        rule = Rule()

        cfn = Template("", {}, regions=["us-east-1"])

        validator = CfnTemplateValidator(True)
        validator = rule.extend_validator(
            validator,
            {},
            create_context_for_template(cfn),
        )

        self.assertEqual(validator.validators["foo"], _validator)
        self.assertIsNotNone(validator.validators["bar"])
        self.assertIsNone(validator.validators["foobar"])

    def test_json_Schema_validator(self):
        class Rule(BaseJsonSchema):
            id = "E3AAAA"

            def __init__(self) -> None:
                super().__init__()
                self.rule_set = {
                    "bar": "E3XXX",
                    "foobar": "E3YYY",
                    "foo": "E3ZZZ",
                }
                self.child_rules = {
                    "E3XXX": RuleWithFunction(),
                    "E3YYY": RuleWithOutFunction(),
                    "E3ZZZ": None,
                }
                self.validators = {
                    "bar": None,
                    "foobar": None,
                    "foo": None,
                }

        rule = Rule()

        validator = CfnTemplateValidator(True)

        def iter_errors(instance):
            yield ValidationError(
                "Failure1", path=deque(["A"]), extra_args={"Foo": "Bar"}
            )
            yield ValidationError(
                "Failure2", path=deque(["A"]), path_override=deque(["Test"])
            )
            yield ValidationError(
                "Failure3", path=deque(["A"]), rule=RuleWithFunction()
            )
            yield ValidationError("Failure4", path=deque(["A"]), validator="foobar")
            yield ValidationError("Failure5", path=deque(["A"]), validator="foo")

        validator.iter_errors = iter_errors

        errors = list(rule.json_schema_validate(validator, {}, []))
        # 5th error is ignored because the rule is None which means
        # the length is 4
        self.assertEqual(len(errors), 4)

        self.assertEqual(errors[0], RuleMatch(["A"], "Failure1"))
        self.assertEqual(errors[0].Foo, "Bar")
        self.assertEqual(errors[0].rule.id, "E3AAAA")
        self.assertEqual(errors[1], RuleMatch(["Test"], "Failure2"))
        self.assertEqual(errors[1].rule.id, "E3AAAA")
        self.assertEqual(errors[2], RuleMatch(["A"], "Failure3"))
        self.assertEqual(errors[2].rule.id, "E3XXX")
        self.assertEqual(errors[3], RuleMatch(["A"], "Failure4"))
        self.assertEqual(errors[3].rule.id, "E3YYY")

    def test_validate(self):
        class ChildRule(CloudFormationLintRule):
            id = "EXXXX"

        class Test(BaseJsonSchema):
            def __init__(self) -> None:
                super().__init__()
                self.rule_set = {"enum": "EXXXX"}
                self.child_rules = {"EXXXX": ChildRule()}

            @property
            def schema(self):
                return {
                    "type": "object",
                    "properties": {"foo": True, "bar": {"enum": ["foobar"]}},
                    "required": ["foo"],
                }

        validator = CfnTemplateValidator(True)
        rule = Test()

        errors = list(rule.validate(validator, {}, {"bar": "bad"}, []))
        self.assertListEqual(
            errors,
            [
                ValidationError(
                    "'foo' is a required property",
                    path=deque([]),
                    rule=Test(),
                    schema_path=deque(["required"]),
                    validator="required",
                    validator_value=["foo"],
                    instance={"bar": "bad"},
                ),
                ValidationError(
                    "'bad' is not one of ['foobar']",
                    path=deque(["bar"]),
                    rule=ChildRule(),
                    schema_path=deque(["properties", "bar", "enum"]),
                    validator="enum",
                    validator_value=["foobar"],
                    instance="bad",
                ),
            ],
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class AccessControlObsolete(CfnLintKeyword):
    """Check if the S3 Bucket Access Controls are configured"""

    id = "W3045"
    shortdesc = "Controlling access to an S3 bucket should be done with bucket policies"
    description = (
        "Nearly all access control configurations can be more successfully achieved "
        "with bucket policies. Consider using bucket policies instead of access "
        "control."
    )
    source_url = "https://docs.aws.amazon.com/AmazonS3/latest/userguide/about-object-ownership.html"
    tags = ["resources", "s3"]

    def __init__(self):
        super().__init__(["Resources/AWS::S3::Bucket/Properties"])

    def validate(
        self, validator: Validator, _, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        property_sets = validator.cfn.get_object_without_conditions(
            instance, ["AccessControl"]
        )

        for property_set in property_sets:
            props = property_set.get("Object")
            access_control = props.get("AccessControl", None)
            if access_control:
                yield ValidationError(
                    (
                        "'AccessControl' is a legacy property. Consider "
                        "using 'AWS::S3::BucketPolicy' instead"
                    ),
                    path=deque(["AccessControl"]),
                    instance=access_control,
                )



"""
Copyright (c) 2013 Julian Berman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

SPDX-License-Identifier: MIT
"""

# Code is taken from jsonschema package and adapted CloudFormation use
# https://github.com/python-jsonschema/jsonschema

from dataclasses import dataclass, field


@dataclass(frozen=True)
class RefResolutionError(Exception):
    """
    A ref could not be resolved.
    """

    cause: str = field(init=True)

    def __str__(self):
        return str(self.cause)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.context import Path, create_context_for_template
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.functions.DynamicReferenceSecretsManagerPath import (
    DynamicReferenceSecretsManagerPath,
)
from cfnlint.template import Template


@pytest.fixture(scope="module")
def rule():
    rule = DynamicReferenceSecretsManagerPath()
    yield rule


@pytest.fixture(scope="module")
def cfn():
    return Template(
        "",
        {},
        regions=["us-east-1"],
    )


@pytest.fixture(scope="module")
def context(cfn):
    return create_context_for_template(cfn)


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "Valid secrets manager",
            "{{resolve:secretsmanager:Parameter}}",
            ["Resources", "MyResource", "Properties", "LoginProfile", "Password"],
            [],
        ),
        (
            "Short list",
            "{{resolve:secretsmanager:Parameter}}",
            ["Parameters", "MyParameter"],
            [
                ValidationError(
                    (
                        "Dynamic reference '{{resolve:secretsmanager:Parameter}}' "
                        "to secrets manager can only be used in resource properties"
                    ),
                    rule=DynamicReferenceSecretsManagerPath(),
                )
            ],
        ),
        (
            "Invalid SSM secure location",
            "{{resolve:secretsmanager:Parameter}}",
            ["Outputs", "MyOutput", "Value"],
            [
                ValidationError(
                    (
                        "Dynamic reference '{{resolve:secretsmanager:Parameter}}' "
                        "to secrets manager can only be used in resource properties"
                    ),
                    rule=DynamicReferenceSecretsManagerPath(),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, path, expected, rule, context, cfn):
    context = context.evolve(
        path=Path(path),
    )
    validator = CfnTemplateValidator(context=context, cfn=cfn)
    errs = list(rule.validate(validator, {"type": "string"}, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.ectwo.Ebs import Ebs


@pytest.fixture(scope="module")
def rule():
    rule = Ebs()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "VolumeType": "gp3",
                "Iops": 3000,
            },
            [],
        ),
        (
            [],  # wrong type
            [],
        ),
        (
            {
                "VolumeType": "io1",
                "Iops": 3000,
            },
            [],
        ),
        (
            {
                "VolumeType": "standard",
            },
            [],
        ),
        (
            {
                "VolumeType": "io1",
            },
            [
                ValidationError(
                    (
                        "'Iops' is a required property when 'VolumeType' "
                        "has a value of 'io1'"
                    ),
                    rule=Ebs(),
                    path=deque([]),
                    validator="required",
                    validator_value=["Iops"],
                    instance={"VolumeType": "io1"},
                    schema={"required": ["Iops"]},
                    schema_path=deque(["allOf", 0, "then", "required"]),
                )
            ],
        ),
        (
            {
                "VolumeType": "standard",
                "Iops": 3000,
            },
            [
                ValidationError(
                    (
                        "Additional properties are not allowed (Iops) was "
                        "unexpected when 'VolumeType' has a value of 'standard'"
                    ),
                    rule=Ebs(),
                    path=deque(["Iops"]),
                    validator=None,
                    validator_value=False,
                    instance=3000,
                    schema=False,
                    schema_path=deque(["allOf", 1, "then", "properties", "Iops"]),
                )
            ],
        ),
    ],
)
def test_backup_lifecycle(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright (c) 2013 Julian Berman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

SPDX-License-Identifier: MIT
"""

# Code is taken from jsonschema package and adapted CloudFormation use
# https://github.com/python-jsonschema/jsonschema
from __future__ import annotations

import heapq
import itertools
from collections import deque
from pprint import pformat
from textwrap import dedent, indent
from typing import TYPE_CHECKING, ClassVar, Deque, Tuple

from cfnlint.jsonschema import _utils

if TYPE_CHECKING:
    from cfnlint.rules import CloudFormationLintRule

WEAK_MATCHES: frozenset[str] = frozenset(["anyOf", "oneOf"])
STRONG_MATCHES: frozenset[str] = frozenset()

_unset = _utils.Unset()


class _Error(Exception):
    _word_for_schema_in_error_message: ClassVar[str]
    _word_for_instance_in_error_message: ClassVar[str]

    def __init__(
        self,
        message: str,
        validator=_unset,
        path=(),
        cause=None,
        context=(),
        validator_value=_unset,
        instance=_unset,
        schema=_unset,
        schema_path=(),
        parent=None,
        type_checker=_unset,
        extra_args=None,
        rule: CloudFormationLintRule | None = None,
        path_override: Deque | Tuple = (),
    ):
        super().__init__(
            message,
            validator,
            path,
            cause,
            context,
            validator_value,
            instance,
            schema,
            schema_path,
            parent,
        )
        self.message = message
        self.path = self.relative_path = deque(path)
        self.schema_path = self.relative_schema_path = deque(schema_path)
        self.context = list(context)
        self.cause = self.__cause__ = cause
        self.validator = validator
        self.validator_value = validator_value
        self.instance = instance
        self.schema = schema
        self.parent = parent
        self._type_checker = type_checker
        self.extra_args = extra_args or {}
        self.rule = rule
        self.path_override = deque(path_override)

        for error in context:
            error.parent = self

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.message!r}>"

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, ValidationError):
            return False

        if self.rule is not None:
            if __value.rule is None:
                return False
            if self.rule.id != __value.rule.id:
                return False
        elif __value.rule is not None:
            return False

        return (
            self.message == __value.message
            and self.path == __value.path
            and self.schema_path == __value.schema_path
            and self.context == __value.context
            and self.cause == __value.cause
            and self.validator == __value.validator
            and self.path_override == __value.path_override
        )

    @property
    def absolute_path(self):
        parent = self.parent
        if parent is None:
            return self.relative_path

        path = deque(self.relative_path)
        path.extendleft(reversed(parent.absolute_path))
        return path

    @property
    def absolute_schema_path(self):
        parent = self.parent
        if parent is None:
            return self.relative_schema_path

        path = deque(self.relative_schema_path)
        path.extendleft(reversed(parent.absolute_schema_path))
        return path

    @property
    def json_path(self):
        path = "$"
        for elem in self.absolute_path:
            if isinstance(elem, int):
                path += "[" + str(elem) + "]"
            else:
                path += "." + elem
        return path

    def _set(self, type_checker=None, **kwargs):
        if type_checker is not None and self._type_checker is _unset:
            self._type_checker = type_checker

        for k, v in kwargs.items():
            if getattr(self, k) is _unset:
                setattr(self, k, v)

    def _contents(self):
        attrs = (
            "message",
            "cause",
            "context",
            "validator",
            "validator_value",
            "path",
            "schema_path",
            "instance",
            "schema",
            "parent",
        )
        return dict((attr, getattr(self, attr)) for attr in attrs)

    def _matches_type(self):
        try:
            expected = self.schema["type"]
        except (KeyError, TypeError):
            return False

        if isinstance(expected, str):
            return self._type_checker.is_type(self.instance, expected)

        return any(
            self._type_checker.is_type(self.instance, expected_type)
            for expected_type in expected
        )


class ValidationError(_Error):
    """
    An instance was invalid under a provided schema.
    """

    _word_for_schema_in_error_message = "schema"
    _word_for_instance_in_error_message = "instance"


class SchemaError(_Error):
    """
    A schema was invalid under its corresponding metaschema.
    """

    _word_for_schema_in_error_message = "metaschema"
    _word_for_instance_in_error_message = "schema"


class UnknownType(Exception):
    """
    A validator was asked to validate an instance against an unknown type.
    """

    def __init__(self, type, instance, schema):
        self.type = type
        self.instance = instance
        self.schema = schema

    def __str__(self):
        prefix = 16 * " "

        return dedent(
            f"""\
            Unknown type {self.type!r} for validator with schema:
                {indent(pformat(self.schema, width=72), prefix).lstrip()}

            While checking instance:
                {indent(pformat(self.instance, width=72), prefix).lstrip()}
            """.rstrip(),
        )


def by_relevance(weak=WEAK_MATCHES, strong=STRONG_MATCHES):
    """
    Create a key function that can be used to sort errors by relevance.

    Arguments:
        weak (set):
            a collection of validation keywords to consider to be
            "weak".  If there are two errors at the same level of the
            instance and one is in the set of weak validation keywords,
            the other error will take priority. By default, :kw:`anyOf`
            and :kw:`oneOf` are considered weak keywords and will be
            superseded by other same-level validation errors.

        strong (set):
            a collection of validation keywords to consider to be
            "strong"
    """

    def relevance(error):
        validator = error.validator
        return (
            -len(error.path),
            validator not in weak,
            validator in strong,
            not error._matches_type(),
        )

    return relevance


relevance = by_relevance()
"""
A key function (e.g. to use with `sorted`) which sorts errors by relevance.

Example:

.. code:: python

    sorted(validator.iter_errors(12), key=jsonschema.exceptions.relevance)
"""


def best_match(errors, key=relevance):
    """
    Try to find an error that appears to be the best match among given errors.

    In general, errors that are higher up in the instance (i.e. for which
    `ValidationError.path` is shorter) are considered better matches,
    since they indicate "more" is wrong with the instance.

    If the resulting match is either :kw:`oneOf` or :kw:`anyOf`, the
    *opposite* assumption is made -- i.e. the deepest error is picked,
    since these keywords only need to match once, and any other errors
    may not be relevant.

    Arguments:
        errors (collections.abc.Iterable):

            the errors to select from. Do not provide a mixture of
            errors from different validation attempts (i.e. from
            different instances or schemas), since it won't produce
            sensical output.

        key (collections.abc.Callable):

            the key to use when sorting errors. See `relevance` and
            transitively `by_relevance` for more details (the default is
            to sort with the defaults of that function). Changing the
            default is only useful if you want to change the function
            that rates errors but still want the error context descent
            done by this function.

    Returns:
        the best matching error, or ``None`` if the iterable was empty

    .. note::

        This function is a heuristic. Its return value may change for a given
        set of inputs from version to version if better heuristics are added.
    """
    errors = iter(errors)
    best = next(errors, None)
    if best is None:
        return
    best = max(itertools.chain([best], errors), key=key)

    while best.context:
        # Calculate the minimum via nsmallest, because we don't recurse if
        # all nested errors have the same relevance (i.e. if min == max == all)
        smallest = heapq.nsmallest(2, best.context, key=key)
        if len(smallest) == 2 and key(smallest[0]) == key(smallest[1]):
            return best
        best = smallest[0]
    return best


class UndefinedTypeCheck(Exception):
    """
    A type checker was asked to check a type it did not have registered.
    """

    def __init__(self, t):
        self.t = t

    def __str__(self):
        return f"Type {self.t!r} is unknown to this type checker"


class FormatError(Exception):
    """
    Validating a format failed.
    """

    def __init__(self, message, cause=None):
        super().__init__(message, cause)
        self.message = message
        self.cause = self.__cause__ = cause

    def __str__(self):
        return self.message



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import unittest
from collections import deque

from cfnlint.context import Context, Path
from cfnlint.context.conditions._conditions import Condition, Conditions


class TestCfnContext(unittest.TestCase):
    def test_class(self):
        context = Context(
            conditions=Conditions.create_from_instance(
                conditions={
                    "Foo": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
                    "Bar": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-west-2"]},
                },
                parameters={},
            ),
        )
        self.assertEqual(context.regions, ["us-east-1"])
        self.assertEqual(
            context.conditions,
            Conditions(
                {
                    "Foo": Condition.create_from_instance(
                        {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]}, {}
                    ),
                    "Bar": Condition.create_from_instance(
                        {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-west-2"]}, {}
                    ),
                },
            ),
        )
        self.assertEqual(context.path, Path())

        # Test evolve
        evolved = context.evolve(
            regions=["us-west-2"],
            conditions=context.conditions.evolve(
                {"Foo": False},
            ),
            path=Path(path=deque([1])),
        )
        self.assertListEqual(evolved.regions, ["us-west-2"])
        self.assertDictEqual(
            evolved.conditions.status,
            {
                "Foo": False,
            },
        )
        self.assertEqual(evolved.path.path, deque([1]))

        evolved_again = evolved.evolve(
            path=evolved.path.descend(path=2),
            conditions=evolved.conditions.evolve(
                {"Bar": True},
            ),
        )
        self.assertListEqual(evolved_again.regions, ["us-west-2"])
        self.assertDictEqual(
            evolved_again.conditions.status, {"Foo": False, "Bar": True}
        )
        self.assertEqual(evolved_again.path.path, deque([1, 2]))
        # no changes to the original
        self.assertListEqual(context.regions, ["us-east-1"])
        self.assertEqual(
            context.conditions,
            Conditions(
                {
                    "Foo": Condition.create_from_instance(
                        {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]}, {}
                    ),
                    "Bar": Condition.create_from_instance(
                        {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-west-2"]}, {}
                    ),
                }
            ),
        )
        self.assertEqual(context.path.path, deque())

        no_path = context.evolve(regions=["us-east-2"])
        self.assertListEqual(no_path.regions, ["us-east-2"])
        self.assertEqual(
            no_path.conditions,
            Conditions(
                {
                    "Foo": Condition.create_from_instance(
                        {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]}, {}
                    ),
                    "Bar": Condition.create_from_instance(
                        {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-west-2"]}, {}
                    ),
                }
            ),
        )
        self.assertEqual(no_path.path.path, deque())

        with self.assertRaises(ValueError):
            evolved_again.evolve(
                conditions=evolved_again.conditions.evolve(
                    {"Foo": True},
                ),
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError

# ruff: noqa: E501
from cfnlint.rules.resources.UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes import (
    UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes,
)


@pytest.fixture(scope="module")
def rule():
    rule = UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "Type": "AWS::EC2::VPC",
            },
            [],
        ),
        (
            {
                "Type": ["AWS::EC2::VPC"],
            },
            [],
        ),
        (
            {
                "Type": "AWS::SQS::Queue",
                "DeletionPolicy": "Foo",
                "UpdateReplacePolicy": "Foo",
            },
            [],
        ),
        (
            {
                "Type": "AWS::SQS::Queue",
            },
            [
                ValidationError(
                    (
                        "'DeletionPolicy' is a required property (The default "
                        "action when replacing/removing a resource is to delete "
                        "it. Set explicit values for stateful resource)"
                    ),
                    rule=UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes(),
                    path=deque([]),
                    validator="required",
                    schema_path=deque(["required"]),
                ),
                ValidationError(
                    (
                        "'UpdateReplacePolicy' is a required property (The default "
                        "action when replacing/removing a resource is to delete "
                        "it. Set explicit values for stateful resource)"
                    ),
                    rule=UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes(),
                    path=deque([]),
                    validator="required",
                    schema_path=deque(["required"]),
                ),
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from cfnlint.decode.node import Mark, dict_node, list_node
from cfnlint.helpers import is_function

if TYPE_CHECKING:
    from cfnlint.context.context import Context


def get_conditions_from_property(instance: Any, is_root: bool = True) -> set[str]:
    """
    Gets the name of the conditions used directly inside the object.

    We do not look at nested objects for conditions.

    Args:
        instance (Any): The object or listto process.
        is_root (bool): If we are at the root of the object. Default: True.

    Returns:
        set[str]: The set of conditions used in the object or list.
    """
    results: set[str] = set()
    if isinstance(instance, list):
        for v in instance:
            results = results.union(get_conditions_from_property(v, is_root=False))
        return results

    fn_k, fn_v = is_function(instance)
    if fn_k == "Fn::If":
        if isinstance(fn_v, list) and len(fn_v) == 3:
            if isinstance(fn_v[0], str):
                results.add(fn_v[0])
                results = results.union(
                    get_conditions_from_property(fn_v[1], is_root=is_root)
                )
                results = results.union(
                    get_conditions_from_property(fn_v[2], is_root=is_root)
                )
        return results
    if is_root:
        if isinstance(instance, dict):
            for k, v in instance.items():
                results = results.union(get_conditions_from_property(v, is_root=False))
    return results


def build_instance_from_scenario(
    instance: Any,
    scenario: dict[str, bool],
    is_root: bool,
    context: "Context",
) -> Any:
    """
    Get object values from a provided scenario.

    This function recursively processes the provided object, resolving any
    conditional logic (such as Fn::If) based on the given scenario.

    Args:
        instance (Any): The object or listto process.
        scenario (dict): The scenario to use when resolving conditional logic.
        is_root (bool): If we are at the root of the object. Default: True.

    Returns:
        dict or list or any: The processed object, with conditional logic resolved.
        The return type can be a dictionary, list, or any other data type,
        depending on the structure of the input object.
    """

    if isinstance(instance, list):
        new_list: list[Any] = list_node(
            [],
            getattr(instance, "start_mark", Mark(0, 0)),
            getattr(instance, "end_mark", Mark(0, 0)),
        )
        for v in instance:
            new_value = build_instance_from_scenario(
                v,
                scenario,
                is_root=False,
                context=context,
            )
            if new_value is not None:
                new_list.append(new_value)
        return new_list

    if isinstance(instance, dict):
        fn_k, fn_v = is_function(instance)
        if fn_k == "Fn::If" and "Fn::If" in context.functions:
            if isinstance(fn_v, list) and len(fn_v) == 3:
                if isinstance(fn_v[0], str):
                    if_path = scenario.get(fn_v[0], None)
                    if if_path is not None:
                        new_value = build_instance_from_scenario(
                            fn_v[1] if if_path else fn_v[2],
                            scenario,
                            is_root,
                            context=context,
                        )
                        if new_value is not None:
                            return new_value
                        return None
            return instance
        if fn_k == "Ref" and fn_v == "AWS::NoValue" and "Ref" in context.functions:
            return None
        if is_root:
            new_obj: dict[str, Any] = dict_node(
                {},
                getattr(instance, "start_mark", Mark(0, 0)),
                getattr(instance, "end_mark", Mark(0, 0)),
            )
            for k, v in instance.items():
                new_value = build_instance_from_scenario(
                    v,
                    scenario,
                    is_root=False,
                    context=context,
                )
                if new_value is not None:
                    new_obj[k] = new_value
            return new_obj

    return instance



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class AnyOf(CloudFormationLintRule):
    """Check Properties Resource Configuration"""

    id = "E3017"
    shortdesc = "Check Properties that need at least one of a list of properties"
    description = (
        "Making sure CloudFormation properties "
        + "that require at least one property from a list. "
        + "More than one can be included."
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint"
    tags = ["resources"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.integration import BaseCliTestCase

from cfnlint import ConfigMixIn


class TestTransformIgnore(BaseCliTestCase):
    """Test Ignoring Transform Failure"""

    scenarios = [
        {
            "filename": (
                "test/fixtures/templates/bad/transform_serverless_template.yaml"
            ),
            "exit_code": 0,
        },
    ]

    def test_templates(self):
        """Test same templates using integration approach"""
        self.run_module_integration_scenarios(
            ConfigMixIn(
                [],
                ignore_checks=["E0001"],
                include_checks=["I"],
                include_experimental=True,
            )
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.helpers import FUNCTIONS
from cfnlint.jsonschema import Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema


class Value(CfnLintJsonSchema):
    """Check if Outputs have string values"""

    id = "E6101"
    shortdesc = "Validate that outputs values are a string"
    description = "Make sure that output values have a type of string"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/outputs-section-structure.html"
    tags = ["outputs"]

    def __init__(self):
        super().__init__(
            keywords=["Outputs/*"],
            all_matches=True,
        )

    def validate(self, validator: Validator, _: Any, instance: Any, schema: Any):
        key = "Value"
        value = instance.get(key)
        if not value:
            return

        conditions = {}
        condition = instance.get("Condition")
        if condition:
            conditions = {condition: True}

        validator = validator.evolve(
            context=validator.context.evolve(
                functions=list(FUNCTIONS),
                conditions=validator.context.conditions.evolve(
                    conditions,
                ),
            )
        )

        for err in validator.descend(
            value,
            schema={
                "type": ["array", "string"],
                "items": {
                    "type": "string",
                },
            },
            path=key,
            property_path=key,
        ):
            if err.rule is None:
                err.rule = self
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import namedtuple

import pytest

from cfnlint import Template
from cfnlint.context.context import create_context_for_template

_Counts = namedtuple("_Counts", ["resources", "parameters", "conditions", "mappings"])


@pytest.mark.parametrize(
    "name,instance,counts",
    [
        (
            "Valid template",
            {
                "Parameters": {
                    "Env": {
                        "Type": "String",
                    }
                },
                "Conditions": {
                    "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]}
                },
                "Mappings": {"Map": {"us-east-1": {"foo": "bar"}}},
                "Resources": {
                    "Bucket": {
                        "Type": "AWS::S3::Bucket",
                    },
                },
            },
            _Counts(resources=1, parameters=1, conditions=1, mappings=1),
        ),
        (
            "Bad types in template",
            {
                "Parameters": [],
                "Conditions": [],
                "Mappings": [],
                "Resources": [],
            },
            _Counts(resources=0, parameters=0, conditions=0, mappings=0),
        ),
        (
            "Invalid type configurations",
            {
                "Parameters": {
                    "BusinessUnit": [],
                    "Env": {
                        "Type": "String",
                    },
                },
                "Mappings": {"AnotherMap": [], "Map": {"us-east-1": {"foo": "bar"}}},
                "Resources": {
                    "Instance": [],
                    "Bucket": {
                        "Type": "AWS::S3::Bucket",
                    },
                },
            },
            _Counts(resources=1, parameters=1, conditions=0, mappings=2),
        ),
        (
            "Invalid mapping second key",
            {
                "Mappings": {
                    "BadKey": {
                        "Foo": [],
                    },
                    "Map": {"us-east-1": {"foo": "bar"}},
                },
            },
            _Counts(resources=0, parameters=0, conditions=0, mappings=2),
        ),
        (
            "Invalid mapping third key",
            {
                "Mappings": {
                    "BadKey": {
                        "Foo": {
                            "Bar": {},
                        },
                    },
                    "Map": {"us-east-1": {"foo": "bar"}},
                },
            },
            _Counts(resources=0, parameters=0, conditions=0, mappings=2),
        ),
    ],
)
def test_create_context(name, instance, counts):
    cfn = Template("", instance, ["us-east-1"])
    context = create_context_for_template(cfn)

    for i in counts._fields:
        if i == "conditions":
            assert len(context.conditions.conditions) == getattr(counts, i), (
                f"Test {name} has {i} {len(getattr(context, i))} "
                f"and expected {getattr(counts, i)}"
            )
        elif i == "mappings":
            assert len(context.mappings.maps) == getattr(counts, i), (
                f"Test {name} has {i} {len(context.mappings.maps)} "
                f"and expected {getattr(counts, i)}"
            )
        else:
            assert len(getattr(context, i)) == getattr(counts, i), (
                f"Test {name} has {i} {len(getattr(context, i))} "
                f"and expected {getattr(counts, i)}"
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class ApproachingMaxLength(CloudFormationLintRule):
    """Check maximum Output name size limit"""

    id = "I6011"
    shortdesc = "Output name limit"
    description = (
        "Check the size of Output names in the template is approaching the upper limit"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["outputs", "limits"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context.context import Path


@pytest.mark.parametrize(
    "name,params,expected",
    [
        (
            "Proper descend",
            {
                "path": "Property",
                "cfn_path": "Property",
                "value_path": "Property",
            },
            Path(
                path=deque(["Property"]),
                cfn_path=deque(["Property"]),
                value_path=deque(["Property"]),
            ),
        ),
        (
            "Exception on using function",
            {
                "path": "Foo",
                "cfn_path": "Fn::If",
            },
            ValueError(),
        ),
        (
            "Exception on using int",
            {
                "path": "Foo",
                "cfn_path": 1,
            },
            ValueError(),
        ),
    ],
)
def test_path_descend(name, params, expected):
    path = Path()
    if isinstance(expected, Exception):
        with pytest.raises(type(expected)):
            path.descend(**params)
    else:
        path = path.descend(**params)
        assert path == expected



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from datetime import date
from typing import Any

from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class PolicyVersion(CfnLintKeyword):
    """Check if IAM Policy Version is correct"""

    id = "W2511"
    shortdesc = "Check IAM Resource Policies syntax"
    description = (
        "See if the elements inside an IAM Resource policy are configured correctly."
    )
    source_url = "https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements.html"
    tags = ["properties", "iam"]

    def __init__(self) -> None:
        super().__init__(["AWS::IAM::Policy/Properties/PolicyDocument/Version"])

    # pylint: disable=unused-argument
    def validate(self, validator: Validator, _, instance: Any, schema):
        if not isinstance(instance, (date, str)):
            return

        if instance in ["2008-10-17", date(2008, 10, 17)]:
            yield ValidationError(
                "IAM Policy Version should be updated to '2012-10-17'",
                rule=self,
            )



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import filecmp
import fnmatch
import json
import logging
import multiprocessing
import os
import re
import shutil
import sys
import zipfile
from copy import copy
from functools import lru_cache
from typing import Any, Dict, Iterator, Sequence

import jsonpatch
import jsonpointer

from cfnlint.helpers import (
    REGION_PRIMARY,
    REGIONS,
    ToPy,
    get_url_retrieve,
    load_resource,
    url_has_newer_version,
)
from cfnlint.schema._exceptions import ResourceNotFoundError
from cfnlint.schema._getatts import AttributeDict
from cfnlint.schema._schema import Schema
from cfnlint.schema.patch import SchemaPatch

LOGGER = logging.getLogger(__name__)


class _FileLocation:
    def __init__(self, path: list[str]):
        self.path_relative = os.path.join(
            os.path.dirname(__file__),
            "..",
            *path,
        )
        self.module = ".".join(["cfnlint"] + path[:])
        self.path = path


class ProviderSchemaManager:
    def __init__(self) -> None:
        self._root = _FileLocation(
            [
                "data",
                "schemas",
                "providers",
            ]
        )
        self._patches = _FileLocation(
            [
                "data",
                "schemas",
                "patches",
            ]
        )
        self._region_primary = ToPy(REGION_PRIMARY)
        self._registry_schemas: dict[str, Schema] = {}
        self._provider_schema_modules: dict[str, Any] = {}
        self.reset()

    def reset(self) -> None:
        """
        Reset's the cache so specs can be reloaded.
        Important function when processing many templates
        and using spec patching
        """
        self._schemas: dict[str, dict[str, Schema]] = {}
        for region in REGIONS:
            self._schemas[region] = {}
        self._removed_types: list[str] = []
        self.get_resource_schema.cache_clear()
        self.get_resource_types.cache_clear()
        self.get_type_getatts.cache_clear()

    def load_registry_schemas(self, path: str) -> None:
        """Load extra registry schemas from a directory

        Args:
            path (str): the directory to load schema files from
        Returns:
            None: None
        """
        for dirpath, _, filenames in os.walk(path):
            for filename in fnmatch.filter(filenames, "*.json"):
                with open(os.path.join(dirpath, filename), "r", encoding="utf-8") as fh:
                    schema = Schema(json.load(fh))
                    self._registry_schemas[schema.type_name] = schema

    def get_resource_schemas_by_regions(
        self, resource_type: str, regions: Sequence[str]
    ) -> Iterator[tuple[list[str], Schema]]:
        """
        Get unique schemas with their associated regions

        Args:
            resource_type (str): the :: version of the resource type
            regions (Sequence[str]): the regions in which we want schems for
        Returns:
            Iterator[tuple[list[str], Schema]]: the unique schemas with
              their associated regions
        """
        cached_regions: list[str] = []
        cached_schema: Schema | None = None
        for region in regions:
            try:
                schema = self.get_resource_schema(region, resource_type)
            except ResourceNotFoundError:
                continue
            if not schema.is_cached and region != REGION_PRIMARY:
                yield [region], schema
            else:
                cached_regions.append(region)
                cached_schema = schema

        if cached_schema is not None:
            yield cached_regions, cached_schema

    def _normalize_resource_type(self, resource_type: str) -> str:
        """
        Normalize the resource type to the correct format

        Args:
            resource_type (str): the :: version of the resource type
        Returns:
            str: the normalized resource type
        """
        if resource_type.startswith("Custom::"):
            resource_type = "AWS::CloudFormation::CustomResource"
        if resource_type.endswith("::MODULE"):
            resource_type = "Module"

        return resource_type

    @lru_cache(maxsize=None)
    def get_resource_schema(self, region: str, resource_type: str) -> Schema:
        """Get the provider resource schema and cache it to speed up future lookups

        Args:
            region (str): the region in which do get the provider schema for
            resource_type (str): the :: version of the resource type
        Returns:
            dict: returns the schema
        """
        if resource_type not in self._registry_schemas:
            resource_type = self._normalize_resource_type(resource_type)

        if resource_type in self._removed_types:
            raise ResourceNotFoundError(resource_type, region)

        reg = ToPy(region)
        rt = ToPy(resource_type)

        schema = self._schemas[reg.name].get(rt.name)
        if schema is not None:
            return schema

        # dynamically import the modules as needed
        self._provider_schema_modules[reg.name] = __import__(
            f"{self._root.module}.{reg.py}", fromlist=[""]
        )
        # check cfn-lint provided schemas
        if rt.name in self._registry_schemas:
            self._schemas[reg.name][rt.name] = self._registry_schemas[rt.name]
            return self._schemas[reg.name][rt.name]

        # load the schema
        if f"{rt.provider}.json" in self._provider_schema_modules[reg.name].cached:
            schema_cached = copy(
                self.get_resource_schema(
                    region=self._region_primary.name,
                    resource_type=rt.name,
                )
            )
            schema_cached.is_cached = True
            self._schemas[reg.name][rt.name] = schema_cached
            return self._schemas[reg.name][rt.name]
        try:
            self._schemas[reg.name][rt.name] = Schema(
                load_resource(
                    self._provider_schema_modules[reg.name],
                    filename=f"{rt.provider}.json",
                )
            )
            return self._schemas[reg.name][rt.name]
        except Exception as e:
            raise ResourceNotFoundError(rt.name, region) from e

    @lru_cache(maxsize=None)
    def get_resource_types(self, region: str) -> list[str]:
        """Get the resource types for a region

        Args:
            region (str): the region in which to get the resource types for
        Returns:
            list[str]: returns a list of resource types
        """
        reg = ToPy(region)

        if self._region_primary.name not in self._provider_schema_modules:
            self._provider_schema_modules[self._region_primary.name] = __import__(
                f"{self._root.module}.{self._region_primary.py}", fromlist=[""]
            )
        resource_types: list[str] = []
        if reg.name not in self._provider_schema_modules:
            self._provider_schema_modules[region] = __import__(
                f"{self._root.module}.{reg.py}", fromlist=[""]
            )
        resource_types.extend(
            rt
            for rt in self._provider_schema_modules[reg.name].types
            if rt not in self._removed_types
        )
        resource_types.extend(list(self._registry_schemas.keys()))

        return resource_types

    def update(self, force: bool) -> None:
        """Update every regions provider schemas

        Args:
            force (bool): force the schemas to be downloaded
        Returns:
            None: returns when complete
        """
        self._update_provider_schema(self._region_primary.name, force=force)
        # pylint: disable=not-context-manager
        with multiprocessing.Pool() as pool:
            # Patch from registry schema
            provider_pool_tuple = [
                (k, force) for k in REGIONS if k != self._region_primary.name
            ]
            pool.starmap(self._update_provider_schema, provider_pool_tuple)

    def _remove_descriptions(self, spec: Any) -> Any:
        if isinstance(spec, dict):
            r: dict[Any, Any] = {}
            for k, v in spec.items():
                if k != "description":
                    r[k] = self._remove_descriptions(v)

            return r
        elif isinstance(spec, list):
            m: list[Any] = []
            for v in spec:
                m.append(self._remove_descriptions(v))

            return m
        else:
            return spec

    def _update_provider_schema(self, region: str, force: bool = False) -> None:
        """Update the provider schemas from the AWS websites

        Args:
            region (str): the region in which do ge the provider schema for
            force (bool): force the schemas to be downloaded
        Returns:
            None: returns when complete
        """
        # China regions in .com.cn
        suffix = ".cn" if region in ["cn-north-1", "cn-northwest-1"] else ""
        url = f"https://schema.cloudformation.{region}.amazonaws.com{suffix}/CloudformationSchema.zip"
        reg = ToPy(region)
        directory = os.path.join(f"{self._root.path_relative}/{reg.py}/")
        directory_pr = os.path.join(
            f"{self._root.path_relative}/{self._region_primary.py}/"
        )

        multiprocessing_logger = multiprocessing.log_to_stderr()

        multiprocessing_logger.debug("Downloading template %s into %s", url, directory)

        # Check to see if we already have the latest version, and if so stop
        if not (url_has_newer_version(url) or force):
            return

        if not os.path.exists(directory):
            os.mkdir(directory)

        try:
            filehandle = get_url_retrieve(url, caching=True)
            # clean folder
            shutil.rmtree(directory)
            with zipfile.ZipFile(filehandle, "r") as zip_ref:
                zip_ref.extractall(directory)

            filenames = [
                f
                for f in os.listdir(directory)
                if os.path.isfile(os.path.join(directory, f)) and f != "__init__.py"
            ]
            # There is no schema for CDK but its an allowable type
            all_types = ["AWS::CDK::Metadata", "Module"]
            with open(f"{directory}module.json", "w", encoding="utf-8") as fh:
                json.dump(
                    {
                        "additionalProperties": True,
                        "type": "object",
                        "typeName": "Module",
                    },
                    fh,
                    indent=1,
                    separators=(",", ": "),
                    sort_keys=True,
                )
                fh.write("\n")
            for filename in filenames:
                with open(f"{directory}{filename}", "r+", encoding="utf-8") as fh:
                    spec = json.load(fh)
                    all_types.append(spec["typeName"])
                    try:
                        spec = self._remove_descriptions(spec)
                        spec = self._patch_provider_schema(spec, filename, "all")
                        spec = self._patch_provider_schema(
                            spec, filename, region=reg.py
                        )
                    except Exception as e:  # pylint: disable=broad-except
                        LOGGER.info(
                            "Issuing patching schema for %s in %s: %s",
                            filename,
                            reg.name,
                            e,
                        )
                    # Back to zero to write spec
                    fh.seek(0)
                    json.dump(
                        spec,
                        fh,
                        indent=1,
                        separators=(",", ": "),
                        sort_keys=True,
                    )
                    fh.write("\n")
                    # Resize doc as needed
                    fh.truncate()

            # if the region is not us-east-1 compare the files to those in us-east-1
            # symlink if the files are the same
            if reg.name != self._region_primary.name:
                cached = ["Module"]
                for filename in os.listdir(directory):
                    if filename != "__init__.py":
                        try:
                            if filecmp.cmp(
                                f"{directory}{filename}",
                                f"{directory_pr}{filename}",
                            ):
                                cached.append(filename)
                                os.remove(f"{directory}{filename}")
                        except FileNotFoundError:
                            pass
                        except Exception as e:  # pylint: disable=broad-except
                            # Exceptions will typically be the file
                            # doesn't exist in primary region
                            LOGGER.info(
                                "Issuing comparing %s into %s: %s",
                                f"{directory}{filename}",
                                f"{directory_pr}{filename}",
                                e,
                            )
                with open(f"{directory}__init__.py", encoding="utf-8", mode="w") as f:
                    f.write("from __future__ import annotations\n\n")
                    f.write("# pylint: disable=too-many-lines\ntypes: list[str] = [\n")
                    for rt in sorted(all_types):
                        f.write(f'    "{rt}",\n')
                    f.write(
                        "]\n\n# pylint: disable=too-many-lines\ncached: list[str] = [\n"
                    )
                    for cache_file in sorted(cached):
                        f.write(f'    "{cache_file}",\n')
                    f.write("]\n")
            else:
                with open(f"{directory}__init__.py", encoding="utf-8", mode="w") as f:
                    f.write("from __future__ import annotations\n\n")
                    f.write("# pylint: disable=too-many-lines\ntypes: list[str] = [\n")
                    for rt in sorted(all_types):
                        f.write(f'    "{rt}",\n')
                    f.write("]\ncached: list[str] = []\n")

        except Exception as e:  # pylint: disable=broad-except
            LOGGER.info("Issuing updating schemas for %s: %s", region, e)

    def _patch_provider_schema(
        self, content: Dict, source_filename: str, region: str
    ) -> Dict:
        """Provides the logic to patch a CloudFormation provider schema file.

        Args:
            content: A Dict representing the data that needs to be patched
            source_filename: The source filename for the JSON patches
            region: The region to apply the patch against
        Returns:
            Dict: returns the patched content
        """
        for patch_type in ["providers", "extensions"]:
            source_dir = source_filename.replace("-", "_").replace(".json", "")
            append_dir = os.path.join(
                self._patches.path_relative, patch_type, region, source_dir
            )
            for dirpath, _, filenames in os.walk(append_dir):
                filenames.sort()
                for filename in fnmatch.filter(filenames, "*.json"):
                    file_path = os.path.basename(filename)
                    module = dirpath.replace(f"{append_dir}", f"{region}").replace(
                        os.path.sep, "."
                    )
                    try:
                        jsonpatch.JsonPatch(
                            load_resource(
                                f"{self._patches.module}.{patch_type}.{module}.{source_dir}",
                                file_path,
                            )
                        ).apply(content, in_place=True)
                    except jsonpatch.JsonPatchConflict as e:
                        LOGGER.info(
                            "Patch already applied %s: %s",
                            os.path.join(append_dir, file_path),
                            str(e),
                        )
                    except jsonpatch.JsonPatchTestFailed as e:
                        LOGGER.info(
                            "Patch test failed %s: %s",
                            os.path.join(append_dir, file_path),
                            str(e),
                        )
                    except jsonpatch.JsonPatchException as e:
                        LOGGER.info(
                            "Patch exception raised for %s: %s",
                            os.path.join(append_dir, file_path),
                            str(e),
                        )
                    except jsonpointer.JsonPointerException as e:
                        LOGGER.info(
                            "Patch exception with pointer %s: %s",
                            os.path.join(append_dir, file_path),
                            str(e),
                        )
                    except Exception as e:  # pylint: disable=broad-exception-caught
                        LOGGER.info(
                            "Unknown exception raised applying patch %s: %s",
                            os.path.join(append_dir, file_path),
                            str(e),
                        )

        return content

    def patch(self, filename: str, regions: Sequence[str]):
        try:
            with open(filename, encoding="utf-8") as fp:
                custom_spec_data = json.load(fp)
                schema_patch = SchemaPatch.from_dict(custom_spec_data)
                for region in regions:
                    self._patch(schema_patch, region)
        except IOError as e:
            if e.errno == 2:
                LOGGER.error("Override spec file not found: %s", filename)
                sys.exit(1)
            elif e.errno == 21:
                LOGGER.error(
                    "Override spec file references a directory, not a file: %s",
                    filename,
                )
                sys.exit(1)
            elif e.errno == 13:
                LOGGER.error(
                    "Permission denied when accessing override spec file: %s", filename
                )
                sys.exit(1)
        except ValueError as err:
            LOGGER.error("Override spec file %s is malformed: %s", filename, err)
            sys.exit(1)

    def _patch(self, patch: SchemaPatch, region: str) -> None:
        """Patch the schemas as needed

        Args:
            patch: The patches to be applied to the schemas
        Returns:
            None: Returns when completed
        """

        resource_types = []
        all_resource_types = self.get_resource_types(region)[:]
        # Remove unsupported resource using includes
        if patch.included_resource_types:
            for include in patch.included_resource_types:
                regex = re.compile(include.replace("*", "(.*)") + "$")
                matches = [
                    string for string in all_resource_types if re.match(regex, string)
                ]

                resource_types.extend(matches)
        else:
            resource_types = all_resource_types[:]

        # Remove unsupported resources using the excludes
        for exclude in patch.excluded_resource_types:
            regex = re.compile(exclude.replace("*", "(.*)") + "$")
            matches = [string for string in resource_types if re.match(regex, string)]
            for match in matches:
                resource_types.remove(match)

        # Remove unsupported resources
        removed_types = list(set(all_resource_types) - set(resource_types))
        if removed_types:
            for removed_type in removed_types:
                self._removed_types.append(removed_type)
            self.get_resource_schema.cache_clear()
            self.get_resource_types.cache_clear()

        for resource_type, patches in patch.patches.items():
            try:
                schema = self.get_resource_schema(
                    resource_type=resource_type, region=region
                )
            except ResourceNotFoundError:
                # Resource type doesn't exist in this region
                continue
            schema.patch(patches=patches)

    @lru_cache(maxsize=None)
    def get_type_getatts(self, resource_type: str, region: str) -> AttributeDict:
        """Get the GetAtts for a type in a region

        Args:
            resource_type: The type of the resource. Example: AWS::S3::Bucket
            region: The region to load the resource type from
        Returns:
            Dict(str, Dict): Returns a Dict where the keys are the attributes and the
                value is the CloudFormation schema description of the attribute
        """
        resource_type = self._normalize_resource_type(resource_type)
        self.get_resource_schema(region=region, resource_type=resource_type)
        return self._schemas[region][resource_type].get_atts


PROVIDER_SCHEMA_MANAGER: ProviderSchemaManager = ProviderSchemaManager()



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema


class CfnLintJsonSchemaRegional(CfnLintJsonSchema):
    def message(self, instance: Any, err: ValidationError) -> str:
        return err.message

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        for region in validator.context.regions:
            region_validator = validator.evolve(
                context=validator.context.evolve(regions=[region]),
                schema=self.schema.get(region, True),
            )
            for err in super()._iter_errors(region_validator, instance):
                err.message = err.message + f" in {region!r}"
                yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema.validators import ValidationError
from cfnlint.rules.resources.ssm.Document import Document


@pytest.fixture(scope="module")
def rule():
    rule = Document()
    yield rule


@pytest.mark.parametrize(
    "name,document,expected",
    [
        (
            "Valid string yaml",
            """
            schemaVersion: "2.2"
            mainSteps:
                - action: aws:runShellScript
            """,
            [],
        ),
        (
            "Valid string json",
            """
            {
                "schemaVersion": "2.2",
                "mainSteps": [
                    {
                        "action": "aws:runShellScript"
                    }
                ]
            }
            """,
            [],
        ),
        (
            "Valid object",
            {
                "schemaVersion": "2.2",
                "mainSteps": [
                    {"action": "aws:runShellScript"},
                ],
            },
            [],
        ),
        (
            "InValid string yaml",
            """
            schemaVersion: 2.2
            mainSteps:
                - action: aws:runShellScript
            """,
            [
                ValidationError(
                    "2.2 is not of type 'string'",
                    rule=Document(),
                    validator="type",
                    schema_path=deque(["properties", "schemaVersion", "type"]),
                    path=deque(["schemaVersion"]),
                )
            ],
        ),
        (
            "Not a valid json or yaml object",
            "arn:aws-us-gov:iam::123456789012:role/test",
            [
                ValidationError(
                    "Document is not of type 'object'",
                    rule=Document(),
                    validator="type",
                    schema_path=deque([]),
                    path=deque([]),
                )
            ],
        ),
        (
            "Invalid schema version in object",
            {
                "schemaVersion": 2.2,
                "mainSteps": [
                    {"action": "aws:runShellScript"},
                ],
            },
            [
                ValidationError(
                    "2.2 is not of type 'string'",
                    rule=Document(),
                    validator="type",
                    schema_path=deque(["properties", "schemaVersion", "type"]),
                    path=deque(["schemaVersion"]),
                )
            ],
        ),
        (
            "Invalid type",
            [],
            [
                ValidationError(
                    "[] is not of type 'object'",
                    rule=Document(),
                    validator="type",
                    schema_path=deque(["type"]),
                    path=deque([]),
                )
            ],
        ),
    ],
)
def test_validate(
    name,
    document,
    expected,
    rule,
    validator,
):
    errs = list(rule.validate(validator, {}, document, {}))

    assert errs == expected



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
import subprocess
import unittest
from copy import deepcopy
from io import StringIO
from pathlib import Path
from typing import Any, Dict, List
from unittest.mock import patch

from cfnlint.config import configure_logging
from cfnlint.decode import cfn_yaml
from cfnlint.formatters import JsonFormatter
from cfnlint.runner import Runner


class BaseCliTestCase(unittest.TestCase):
    """Used for Testing CLI"""

    scenarios: List[Dict[str, Any]] = []

    def setUp(self):
        """Common Settings"""
        self.maxDiff = None

    def run_scenarios(self, extra_params=None):
        """Success test"""
        extra_params = extra_params or []
        for scenario in self.scenarios:
            filename = scenario.get("filename")
            results_filename = scenario.get("results_filename")
            expected_results = scenario.get("results", [])
            exit_code = scenario.get("exit_code")

            if results_filename and not expected_results:
                with open(results_filename, encoding="utf-8") as json_data:
                    expected_results = json.load(json_data)

            for result in expected_results:
                result["Filename"] = str(Path(result.get("Filename")))

            try:
                result = subprocess.check_output(
                    ["cfn-lint"] + extra_params + ["--format", "json", "--", filename]
                )

                self.assertEqual(
                    0,
                    exit_code,
                    "Expected {} exit code, got {} on {}".format(
                        exit_code, 0, filename
                    ),
                )

                if isinstance(result, bytes):
                    result = result.decode("utf8")
                matches = json.loads(result)

                self.assertCountEqual(
                    expected_results,
                    matches,
                    "Expected {} failures, got {} on {}".format(
                        len(expected_results), len(matches), filename
                    ),
                )

            except subprocess.CalledProcessError as error:
                self.assertEqual(
                    error.returncode,
                    exit_code,
                    "Expected {} exit code, got {} on {}".format(
                        exit_code, error.returncode, filename
                    ),
                )

                if isinstance(error.output, bytes):
                    error.output = error.output.decode("utf8")
                matches = json.loads(error.output)

                self.assertCountEqual(
                    expected_results,
                    matches,
                    "Expected {} failures, got {} on {}".format(
                        len(expected_results), len(matches), filename
                    ),
                )

    def run_module_integration_scenarios(self, config):
        """Test using cfnlint as a module integrated into another package"""

        configure_logging(None, False)

        for scenario in self.scenarios:
            filename = scenario.get("filename")
            results_filename = scenario.get("results_filename")
            expected_results = scenario.get("results", [])

            if results_filename and not expected_results:
                with open(results_filename, encoding="utf-8") as json_data:
                    expected_results = json.load(json_data)

            for result in expected_results:
                result["Filename"] = str(Path(result.get("Filename")))

            # template = cfn_yaml.load(filename)
            scenario_config = deepcopy(config)
            scenario_config.cli_args.template_alt = [filename]
            scenario_config.cli_args.format = "json"

            runner = Runner(scenario_config)

            print(f"Running test for {filename!r}")
            with patch("sys.exit") as exit:
                with patch("sys.stdout", new=StringIO()) as out:
                    runner.cli()
                    exit.assert_called_once_with(scenario.get("exit_code", 0))

                    output = json.loads(out.getvalue())
                    self.assertEqual(
                        expected_results,
                        output,
                        f"Test for {filename!r} got results: {output!r}",
                    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import deque
from dataclasses import InitVar, dataclass, field, fields
from typing import Any, Deque, Iterator, Sequence, Set, Tuple

from cfnlint.context._mappings import Mappings
from cfnlint.context.conditions._conditions import Conditions
from cfnlint.helpers import (
    BOOLEAN_STRINGS_TRUE,
    FUNCTIONS,
    PSEUDOPARAMS,
    REGION_PRIMARY,
    TRANSFORM_SAM,
)
from cfnlint.schema import PROVIDER_SCHEMA_MANAGER, AttributeDict

_PSEUDOPARAMS_NON_REGION = ["AWS::AccountId", "AWS::NoValue", "AWS::StackName"]


@dataclass
class Transforms:
    # Template level parameters
    transforms: InitVar[str | list[str] | None]
    _transforms: list[str] = field(init=False, default_factory=list)

    def __post_init__(self, transforms) -> None:
        if transforms is None:
            return
        if not isinstance(transforms, list):
            transforms = [transforms]

        for transform in transforms:
            if not isinstance(transform, str):
                continue
            self._transforms.append(transform)

    def has_language_extensions_transform(self):
        lang_extensions_transform = "AWS::LanguageExtensions"
        return bool(lang_extensions_transform in self._transforms)

    def has_sam_transform(self):
        return bool(TRANSFORM_SAM in self._transforms)


@dataclass(frozen=True)
class Path:
    """
    A `Path` keeps track of the different Path values
    """

    # path keeps track of the path as we move down the template
    # Example: Resources, MyResource, Properties, Name, ...
    path: Deque[str | int] = field(init=True, default_factory=deque)

    # value_path is an override of the value if we got it from another place
    # like a Parameter default value
    # Example: Parameters, MyParameter, Default, ...
    value_path: Deque[str | int] = field(init=True, default_factory=deque)

    # cfn_path is a generic path used by cfn-lint to help make
    # writing rules easier.  The resource name is replaced by the type
    # lists are replaced with a *
    # Example: Resources, AWS::S3::Bucket, Properties, Name, ...
    # Example: Resources, *, Type
    cfn_path: Deque[str] = field(init=True, default_factory=deque)

    def descend(self, **kwargs) -> "Path":
        """
        Create a new Path by appending values
        """
        cls = self.__class__

        if kwargs.get("cfn_path") in FUNCTIONS:
            raise ValueError(f"{kwargs['cfn_path']!r} cannot be in 'cfn_path'")
        elif isinstance(kwargs.get("cfn_path"), int):
            raise ValueError(f"{kwargs['cfn_path']!r} cannot be an integer")

        for f in fields(Path):
            if kwargs.get(f.name) is not None:
                kwargs[f.name] = getattr(self, f.name) + deque([kwargs[f.name]])
            else:
                kwargs[f.name] = getattr(self, f.name)

        return cls(**kwargs)

    def evolve(self, **kwargs):
        """
        Create a new path without appending values
        """
        cls = self.__class__

        for f in fields(Path):
            kwargs.setdefault(f.name, getattr(self, f.name))

        return cls(**kwargs)

    @property
    def path_string(self):
        return "/".join(str(p) for p in self.path)

    @property
    def cfn_path_string(self):
        return "/".join(self.cfn_path)


@dataclass(frozen=True)
class Context:
    """
    A `Context` keeps track of the current context that we are evaluating against
    Arguments:

        region:

            The region being evaluated against.

        conditions:

            The conditions being used and their current state
    """

    # what regions we are processing
    regions: Sequence[str] = field(
        init=True, default_factory=lambda: list([REGION_PRIMARY])
    )

    # supported functions at this point in the template
    functions: Sequence[str] = field(init=True, default_factory=list)

    path: Path = field(init=True, default_factory=Path)

    # cfn-lint Template class
    parameters: dict[str, "Parameter"] = field(init=True, default_factory=dict)
    resources: dict[str, "Resource"] = field(init=True, default_factory=dict)
    conditions: Conditions = field(init=True, default_factory=Conditions)
    mappings: Mappings = field(init=True, default_factory=Mappings)

    strict_types: bool = field(init=True, default=True)

    pseudo_parameters: Set[str] = field(
        init=True, default_factory=lambda: set(PSEUDOPARAMS)
    )

    # Combiniation of storing any resolved ref
    # and adds in any Refs available from things like Fn::Sub
    ref_values: dict[str, Any] = field(init=True, default_factory=dict)

    transforms: Transforms = field(init=True, default_factory=lambda: Transforms([]))

    # is the value a resolved value
    is_resolved_value: bool = field(init=True, default=False)

    def evolve(self, **kwargs) -> "Context":
        """
        Create a new context without merging together attributes
        """
        cls = self.__class__

        if "ref_values" in kwargs:
            new_ref_values = self.ref_values.copy()
            new_ref_values.update(kwargs["ref_values"])
            kwargs["ref_values"] = new_ref_values

        for f in fields(Context):
            if f.init:
                kwargs.setdefault(f.name, getattr(self, f.name))

        return cls(**kwargs)

    def ref_value(self, instance: str) -> Iterator[Tuple[str | list[str], "Context"]]:
        if instance in PSEUDOPARAMS and instance not in self.pseudo_parameters:
            return

        if instance in self.ref_values:
            yield self.ref_values[instance], self
            return

        # Non regionalized items first
        if instance in _PSEUDOPARAMS_NON_REGION and instance in self.pseudo_parameters:
            pseudo_value = _get_pseudo_value(instance)
            if pseudo_value is not None:
                yield pseudo_value, self.evolve(ref_values={instance: pseudo_value})
            return
        if instance in self.parameters:
            for v, path in self.parameters[instance].ref(self):

                # validate that ref is possible with path
                # need to evaluate if Fn::If would be not true if value is
                # what it is
                yield v, self.evolve(
                    path=self.path.evolve(
                        value_path=deque(["Parameters", instance]) + path
                    ),
                    ref_values={instance: v},
                )
            return

        # Regionalized values second
        if instance in PSEUDOPARAMS and instance in self.pseudo_parameters:
            for region in self.regions:
                # We can resolve all region based pseudo values
                # as we are now deciding on a region.
                yield _get_pseudo_value_by_region(instance, region), self.evolve(
                    regions=[region],
                    ref_values={
                        p: _get_pseudo_value_by_region(p, region)
                        for p in PSEUDOPARAMS
                        if p not in _PSEUDOPARAMS_NON_REGION
                    },
                )

    @property
    def refs(self):
        pseudo_params = list(self.pseudo_parameters)
        pseudo_params.sort()
        return (
            list(self.parameters.keys())
            + list(self.resources.keys())
            + pseudo_params
            + list(self.ref_values.keys())
        )


def _get_pseudo_value(parameter: str) -> str | list[str] | None:
    if parameter == "AWS::AccountId":
        return "123456789012"
    if parameter == "AWS::StackName":
        return "teststack"
    return None


def _get_pseudo_value_by_region(parameter: str, region: str) -> str | list[str]:
    if parameter == "AWS::NotificationARNs":
        return [f"arn:{_get_partition(region)}:sns:{region}:123456789012:notification"]
    if parameter == "AWS::Partition":
        return _get_partition(region)
    if parameter == "AWS::Region":
        return region
    if parameter == "AWS::StackId":
        return (
            f"arn:{_get_partition(region)}:cloudformation:{region}"
            ":123456789012:stack/teststack/51af3dc0-da77-11e4-872e-1234567db123"
        )
    if region in ("cn-north-1", "cn-northwest-1"):
        return "amazonaws.com.cn"
    else:
        return "amazonaws.com"


def _get_partition(region) -> str:
    if region in ("us-gov-east-1", "us-gov-west-1"):
        return "aws-us-gov"
    if region in ("cn-north-1", "cn-northwest-1"):
        return "aws-cn"
    else:
        return "aws"


class _Ref(ABC):
    """
    Abstract class for all ref types
    """

    @abstractmethod
    def ref(self, context: Context) -> Iterator[Any]:
        pass


@dataclass
class Parameter(_Ref):
    """
    This class holds a parameter and its attributes
    """

    type: str = field(init=False)
    default: Any = field(init=False)
    allowed_values: Any = field(init=False)
    description: str | None = field(init=False)

    parameter: InitVar[Any]

    def __post_init__(self, parameter) -> None:
        if not isinstance(parameter, dict):
            raise ValueError("Parameter must be a object")
        self.default = None
        self.allowed_values = []
        self.min_value = None
        self.max_value = None
        self.no_echo = False

        t = parameter.get("Type")
        if not isinstance(t, str):
            raise ValueError("Type must be a string")
        self.type = t

        self.description = parameter.get("Description")

        # SSM Parameter defaults and allowed values point to
        # SSM paths not to the actual values
        if self.type.startswith("AWS::SSM::Parameter::"):
            return

        if self.type == "CommaDelimitedList" or self.type.startswith("List<"):
            if "Default" in parameter:
                default = parameter.get("Default", "")
                if isinstance(default, str):
                    self.default = default.split(",")
                else:
                    self.default = [default]
            for allowed_value in parameter.get("AllowedValues", []):
                if isinstance(allowed_value, str):
                    self.allowed_values.append(allowed_value.split(","))
                else:
                    self.allowed_values.append([allowed_value])
        else:
            self.default = parameter.get("Default")
            self.allowed_values = parameter.get("AllowedValues")

        self.min_value = parameter.get("MinValue")
        self.max_value = parameter.get("MaxValue")
        if parameter.get("NoEcho") in list(BOOLEAN_STRINGS_TRUE) + [True]:
            self.no_echo = True

    def ref(self, context: Context) -> Iterator[Tuple[Any, deque]]:
        if self.allowed_values:
            for i, allowed_value in enumerate(self.allowed_values):
                if isinstance(allowed_value, list):
                    yield [str(x) for x in allowed_value], deque(["AllowedValues", i])
                else:
                    yield str(allowed_value), deque(["AllowedValues", i])
            # assume default is an allowed value so we skip it
            return

        if self.default is not None:
            if isinstance(self.default, list):
                yield [str(x) for x in self.default], deque(["Default"])
            else:
                yield str(self.default), deque(["Default"])

        if self.min_value is not None:
            yield str(self.min_value), deque(["MinValue"])

        if self.max_value is not None:
            yield str(self.max_value), deque(["MaxValue"])


@dataclass
class Resource(_Ref):
    """
    This class holds a resources and its type
    """

    type: str = field(init=False)
    condition: str | None = field(init=False, default=None)
    resource: InitVar[Any]

    def __post_init__(self, resource) -> None:
        if not isinstance(resource, dict):
            raise ValueError("Resource must be a object")
        t = resource.get("Type")
        if not isinstance(t, str):
            raise ValueError("Type must be a string")
        self.type = t
        if self.type.startswith("Custom::"):
            self.type = "AWS::CloudFormation::CustomResource"

        c = resource.get("Condition")
        if not isinstance(t, str):
            raise ValueError("Condition must be a string")
        self.condition = c

    def get_atts(self, region: str = "us-east-1") -> AttributeDict:
        return PROVIDER_SCHEMA_MANAGER.get_type_getatts(self.type, region)

    def ref(self, context: Context) -> Iterator[Any]:
        return
        yield


def _init_parameters(parameters: Any) -> dict[str, Parameter]:
    obj = {}
    if not isinstance(parameters, dict):
        raise ValueError("Parameters must be a object")
    for k, v in parameters.items():
        try:
            obj[k] = Parameter(v)
        except ValueError:
            pass

    return obj


def _init_resources(resources: Any) -> dict[str, Resource]:
    obj = {}
    if not isinstance(resources, dict):
        raise ValueError("Resource must be a object")
    for k, v in resources.items():
        try:
            obj[k] = Resource(v)
        except ValueError:
            pass
    return obj


def _init_transforms(transforms: Any) -> Transforms:
    if isinstance(transforms, (str, list)):
        return Transforms(transforms)
    return Transforms([])


def create_context_for_template(cfn):
    parameters = {}
    try:
        parameters = _init_parameters(cfn.template.get("Parameters", {}))
    except (ValueError, AttributeError):
        pass

    resources = {}
    try:
        resources = _init_resources(cfn.template.get("Resources", {}))
    except (ValueError, AttributeError):
        pass

    transforms = _init_transforms(cfn.template.get("Transform", []))

    try:
        conditions = Conditions.create_from_instance(
            cfn.template.get("Conditions", {}), parameters
        )
    except (ValueError, AttributeError):
        conditions = Conditions.create_from_instance({}, {})

    mappings = Mappings.create_from_dict(cfn.template.get("Mappings", {}))

    return Context(
        parameters=parameters,
        resources=resources,
        conditions=conditions,
        transforms=transforms,
        mappings=mappings,
        regions=cfn.regions,
        path=Path(),
        functions=["Fn::Transform"],
    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque

import pytest

from cfnlint.context import Context
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema

_schema = {
    "type": "object",
    "properties": {
        "a": {
            "type": "string",
        },
        "b": {"type": "object", "properties": {"1": {"type": "string"}}},
        "c": {
            "type": "string",
            "pattern": "^foo$",
        },
    },
}


class _Error(CfnLintJsonSchema):
    id = "EXXXX"
    shortdesc = "Rule to create errors"
    description = "Rule to create errors"

    def __init__(self) -> None:
        super().__init__([], None, True)


class _BestError(CfnLintJsonSchema):
    id = "EXXXX"
    shortdesc = "Rule that returns the best error"
    description = "A longer description"

    def __init__(self) -> None:
        super().__init__([], None, False)


@pytest.mark.parametrize(
    "name,rule,instance,expected_errs",
    [
        (
            "All matches returned",
            _Error(),
            {"a": [], "b": []},
            [
                ValidationError(
                    message="[] is not of type 'string'",
                    rule=_Error(),
                    validator="type",
                    path=deque(["a"]),
                    schema_path=deque(["properties", "a", "type"]),
                ),
                ValidationError(
                    message="[] is not of type 'object'",
                    rule=_Error(),
                    validator="type",
                    path=deque(["b"]),
                    schema_path=deque(["properties", "b", "type"]),
                ),
            ],
        ),
        (
            "Best match",
            _BestError(),
            {"a": [], "b": []},
            [
                ValidationError(
                    message="Rule that returns the best error",
                    validator="type",
                    path=deque(["a"]),
                    rule=_BestError(),
                    schema_path=deque(["properties", "a", "type"]),
                ),
            ],
        ),
        (
            "Validation error with dynamic reference",
            _BestError(),
            {"c": "{{resolve:ssm:/SQS_Queue/SQS_ARN}}"},
            [
                ValidationError(
                    message="Rule that returns the best error",
                    validator="pattern",
                    path=deque(["c"]),
                    rule=_BestError(),
                    schema_path=deque(["properties", "c", "pattern"]),
                ),
            ],
        ),
    ],
)
def test_cfn_schema(name, rule, instance, expected_errs):
    context = Context(regions=["us-east-1"])
    validator = CfnTemplateValidator(schema=_schema, context=context)

    errs = list(rule.validate(validator, {}, instance, _schema))
    assert errs == expected_errs



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.context.conditions._equals import _equals_cmp_key


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid",
            [
                "A",
                "B",
                {"Ref": "A"},
                {"Ref": "B"},
                {"Key": "Value"},
                {"Fn::FindInMap": ["A", "B", "C"]},
                {"Fn::FindInMap": ["X", "Y", "Z"]},
            ],
            [
                "A",
                "B",
                {"Fn::FindInMap": ["X", "Y", "Z"]},
                {"Fn::FindInMap": ["A", "B", "C"]},
                {"Ref": "A"},
                {"Ref": "B"},
                {"Key": "Value"},
            ],
        ),
        (
            "Different order same results",
            [
                {"Fn::FindInMap": ["A", "B", "C"]},
                {"Fn::FindInMap": ["X", "Y", "Z"]},
                {"Ref": "B"},
                {"Ref": "A"},
                {"Key": "Value"},
                "B",
                "A",
            ],
            [
                "A",
                "B",
                {"Fn::FindInMap": ["X", "Y", "Z"]},
                {"Fn::FindInMap": ["A", "B", "C"]},
                {"Ref": "A"},
                {"Ref": "B"},
                {"Key": "Value"},
            ],
        ),
        (
            "Two Refs",
            [
                {"Ref": "B"},
                {"Ref": "A"},
            ],
            [
                {"Ref": "A"},
                {"Ref": "B"},
            ],
        ),
        (
            "Ref with non mapping",
            [
                {"Ref": "A"},
                {"Key": "Value"},
            ],
            [
                {"Ref": "A"},
                {"Key": "Value"},
            ],
        ),
        (
            "Ref with non mapping backwards",
            [
                {"Key": "Value"},
                {"Ref": "A"},
            ],
            [
                {"Ref": "A"},
                {"Key": "Value"},
            ],
        ),
        (
            "None functions",
            [
                {"Key": "Foo"},
                {"Key": "Bar"},
            ],
            [
                {"Key": "Foo"},
                {"Key": "Bar"},
            ],
        ),
    ],
)
def test_sorted(name, instance, expected):
    assert expected == sorted(
        instance, key=_equals_cmp_key
    ), f"{name} got {sorted(instance, key=_equals_cmp_key)}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import unittest
from collections import namedtuple
from typing import List

from cfnlint.context import Context
from cfnlint.jsonschema._keywords_cfn import FnItems, cfn_type
from cfnlint.jsonschema.exceptions import UnknownType
from cfnlint.jsonschema.validators import CfnTemplateValidator


class _Fns:
    def __init__(self, resp) -> None:
        self.resp = resp

    def get_value(self, instance, region):
        for item in self.resp:
            if isinstance(item, Exception):
                raise item
            yield item


class _Cfn:
    def __init__(self, resp) -> None:
        self.functions = _Fns(resp)


_T = namedtuple("_T", ["name", "instance", "schema", "errors", "cfn_response"])


class Base(unittest.TestCase):
    def message_errors(self, name, instance, errors, schema, **kwargs):
        cfn_validator = CfnTemplateValidator({}).extend(validators={"type": cfn_type})
        context = Context(regions=["us-east-1"], strict_types=False)
        cls = kwargs.pop("cls", cfn_validator(schema=schema, context=context))
        validator = cls.evolve(**kwargs)
        i_errors = list(validator.iter_errors(instance))
        self.assertEqual(
            len(errors),
            len(i_errors),
            msg=(
                f"{name}: Expected exactly {len(errors)} error, "
                f"found {i_errors!r}, need {errors!r}"
            ),
        )
        err_messages = [err.message for err in i_errors]
        for err in errors:
            self.assertIn(
                err, err_messages, msg=f"{name}: expected {err} to be in {i_errors!r}"
            )

    def iterate_tests(self, tests: List[_T]):
        for test in tests:
            self.message_errors(
                name=test.name,
                instance=test.instance,
                schema=test.schema,
                errors=test.errors,
                cfn=_Cfn(test.cfn_response),
            )


class TestCfnTypes(Base):
    def build_execute_tests(self, instance, supported_types) -> None:
        _all_types = ["string", "number", "integer", "boolean", "array", "object"]
        tests: List[_T] = []
        for t in _all_types:
            reprs = ", ".join(repr(type) for type in [t])
            if t in supported_types:
                tests.append(
                    _T(
                        f"{instance} can return {t}",
                        instance,
                        {"type": t},
                        [],
                        [],
                    )
                )
            else:
                tests.append(
                    _T(
                        f"{instance} cannot return {t}",
                        instance,
                        {"type": t},
                        [f"{instance!r} is not of type {reprs}"],
                        [],
                    )
                )
        self.iterate_tests(tests)

    def test_standard_values(self):
        self.build_execute_tests({"Foo": True, "Bar": True}, ["object"])
        self.build_execute_tests({"Foo": True}, ["object"])
        self.build_execute_tests("a", ["string"])
        self.build_execute_tests("1.5", ["string", "number"])
        self.build_execute_tests("1", ["string", "number", "integer"])
        self.build_execute_tests(True, ["string", "boolean"])
        self.build_execute_tests("true", ["string", "boolean"])

    def test_null_values(self):
        self.message_errors(
            "Null value with string type",
            None,
            ["None is not of type 'string'"],
            {"type": "string"},
        )
        self.message_errors(
            "String value with null type",
            "foo",
            ["'foo' is not of type 'null'"],
            {"type": "null"},
        )
        self.message_errors(
            "Object value with null type",
            {},
            ["{} is not of type 'null'"],
            {"type": "null"},
        )
        self.message_errors(
            "None value with multiple types",
            None,
            [],
            {"type": ["string", "null"]},
        )


class TestMultiCfnTypes(Base):
    def build_execute_tests(self, instance, supported_types, unsupported_type) -> None:
        tests: List[_T] = []
        check_types = supported_types + [unsupported_type]
        tests.append(
            _T(
                f"{instance} can return one of {check_types!r}",
                instance,
                {"type": check_types},
                [],
                [],
            )
        )

        self.iterate_tests(tests)

    def test_standard_values(self):
        # all these checks should return true as there is one good value
        self.build_execute_tests({"Foo": True, "Bar": True}, ["object"], "string")
        self.build_execute_tests({"Foo": True}, ["object"], "string")
        self.build_execute_tests("a", ["string"], "number")
        self.build_execute_tests("1.5", ["string", "number"], "integer")
        self.build_execute_tests("1", ["string", "number", "integer"], "boolean")
        self.build_execute_tests(True, ["string", "boolean"], "number")
        self.build_execute_tests("true", ["string", "boolean"], "object")


class TestCfnTypeFailure(Base):
    def test_cfn_type_failure(self):
        with self.assertRaises(UnknownType) as err:
            CfnTemplateValidator({}).is_type("foo", "bar")
        self.assertIn(
            "Unknown type 'bar' for validator with schema", str(err.exception)
        )


class TestFnItems(Base):

    def test_change_type(self):
        schema = [
            {
                "schema": {"type": "string"},
            },
        ]
        validator = CfnTemplateValidator({})

        items = list(FnItems().validate(validator, schema, [1], {}))

        self.assertListEqual(items, [])



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.parameters.Used import Used  # pylint: disable=E0401


class TestParameterUsed(BaseRuleTestCase):
    """Test template parameter configurations"""

    def setUp(self):
        """Setup"""
        super(TestParameterUsed, self).setUp()
        self.collection.register(Used())
        self.success_templates = [
            "test/fixtures/templates/good/parameters/used_transforms.yaml",
            "test/fixtures/templates/good/parameters/used_transform_removed.yaml",
            "test/fixtures/templates/good/parameters/used_transform_language_extension.json",
            "test/fixtures/templates/good/parameters/not_used_parameters.yaml",
        ]

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/parameters/configuration.yaml", 17
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema import CfnLintKeyword
from cfnlint.schema import PROVIDER_SCHEMA_MANAGER


class GetAttFormat(CfnLintKeyword):
    id = "E1040"
    shortdesc = "Check if GetAtt matches destination format"
    description = (
        "Validate that if source and destination format exists that they match"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/best-practices.html#parmtypes"
    tags = ["parameters", "ec2", "imageid"]

    def __init__(self):
        super().__init__(["*"])
        self.parent_rules = ["E1010"]
        self._exceptions = [
            # Need to measure for completeness of automation
            "AWS::EC2::SecurityGroup.GroupId",
            "AWS::EC2::SecurityGroup.GroupIds",
        ]
        self._resource_type_exceptions = [
            "AWS::CloudFormation::CustomResource",
            "AWS::CloudFormation::Stack",
            "AWS::ServiceCatalog::CloudFormationProvisionedProduct",
        ]

        self._resource_type_attribute_exceptions = [("AWS::SSM::Parameter", "Value")]

    def validate(
        self, validator: Validator, _, instance: Any, schema: Any
    ) -> ValidationResult:
        fmt = schema.get("format")
        if not fmt or fmt in self._exceptions:
            return

        resource, attr = instance[0:2]

        t = validator.context.resources[resource].type
        for (
            regions,
            resource_schema,
        ) in PROVIDER_SCHEMA_MANAGER.get_resource_schemas_by_regions(
            t, validator.context.regions
        ):
            region = regions[0]

            if t in self._resource_type_exceptions:
                return

            if (t, attr) in self._resource_type_attribute_exceptions:
                return

            getatt_ptr = validator.context.resources[resource].get_atts(region)[attr]

            getatt_schema = resource_schema.resolver.resolve_cfn_pointer(getatt_ptr)
            getatt_fmt = getatt_schema.get("format")
            if getatt_fmt != fmt:
                yield ValidationError(
                    f"{{'Fn::GetAtt': {instance!r}}} that does not match {fmt!r}",
                    rule=self,
                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

# pylint: disable=cyclic-import
import json

import cfnlint.rules.custom.Operators


# pylint: disable=too-many-return-statements
def make_rule(line, lineNumber):
    """Object Maker Function"""

    def get_value(value):
        raw_value = value.strip()
        try:
            return json.loads(raw_value)
        except ValueError:
            return raw_value

    def set_arguments(argument, error):
        raw_value = argument.split(error)
        value = process_sets(raw_value[0])
        error_message = None
        if len(raw_value) > 1:
            error_message = raw_value[1].strip().strip('"')
        return value, error_message

    def process_sets(raw_value):
        if (
            len(raw_value) > 1
            and raw_value[0] == "["
            and (raw_value[-2] == "]" or raw_value[-1] == "]")
        ):
            raw_value = raw_value[1:-2]
            raw_value = raw_value.split(",")
            for x in raw_value:
                x = x.strip()
            return raw_value

        return raw_value

    line = line.rstrip()
    # check line is not a comment or empty line
    if not line.startswith("#") and line != "":
        rule_id = lineNumber + 9000
        line = line.split(" ", 3)
        error_level = "E"
        if len(line) == 4:
            resourceType = line[0]
            prop = line[1]
            operator = line[2]
            value = None
            error_message = None
            if "WARN" in line[3]:
                error_level = "W"
                value, error_message = set_arguments(line[3], "WARN")
            elif "ERROR" in line[3]:
                error_level = "E"
                value, error_message = set_arguments(line[3], "ERROR")
            else:
                value = process_sets(line[3])
                value = get_value(line[3])

            if isinstance(value, str):
                value = value.strip().strip('"')

            if operator in ["EQUALS", "=="]:
                return cfnlint.rules.custom.Operators.CreateEqualsRule(
                    error_level + str(rule_id), resourceType, prop, value, error_message
                )
            if operator in ["NOT_EQUALS", "!="]:
                return cfnlint.rules.custom.Operators.CreateNotEqualsRule(
                    error_level + str(rule_id), resourceType, prop, value, error_message
                )
            if operator == "REGEX_MATCH":
                return cfnlint.rules.custom.Operators.CreateRegexMatchRule(
                    error_level + str(rule_id), resourceType, prop, value, error_message
                )
            if operator == "IN":
                return cfnlint.rules.custom.Operators.CreateInSetRule(
                    error_level + str(rule_id), resourceType, prop, value, error_message
                )
            if operator == "NOT_IN":
                return cfnlint.rules.custom.Operators.CreateNotInSetRule(
                    error_level + str(rule_id), resourceType, prop, value, error_message
                )
            if operator == ">":
                return cfnlint.rules.custom.Operators.CreateGreaterRule(
                    error_level + str(rule_id), resourceType, prop, value, error_message
                )
            if operator == ">=":
                return cfnlint.rules.custom.Operators.CreateGreaterEqualRule(
                    error_level + str(rule_id), resourceType, prop, value, error_message
                )
            if operator == "<":
                return cfnlint.rules.custom.Operators.CreateLesserRule(
                    error_level + str(rule_id), resourceType, prop, value, error_message
                )
            if operator == "<=":
                return cfnlint.rules.custom.Operators.CreateLesserEqualRule(
                    error_level + str(rule_id), resourceType, prop, value, error_message
                )
            if operator == "IS":
                if value in ["DEFINED", "NOT_DEFINED"]:
                    return cfnlint.rules.custom.Operators.CreateCustomIsDefinedRule(
                        error_level + str(rule_id),
                        resourceType,
                        prop,
                        value,
                        error_message,
                    )
                return cfnlint.rules.custom.Operators.CreateInvalidRule(
                    "E" + str(rule_id), f"{operator} {value}"
                )

        return cfnlint.rules.custom.Operators.CreateInvalidRule(
            "E" + str(rule_id), operator
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.other.ssm
from cfnlint.decode import decode_str
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails
from cfnlint.schema.resolver import RefResolver


class Document(CfnLintJsonSchema):
    id = "E3051"
    shortdesc = "Validate the structure of a SSM document"
    description = (
        "SSM documents are nested JSON/YAML in CloudFormation "
        "this rule adds validation to those documents"
    )
    source_url = "https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements.html"
    tags = ["properties", "ssm", "document"]

    def __init__(self):
        super().__init__(
            ["Resources/AWS::SSM::Document/Properties/Content"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.other.ssm,
                filename="document.json",
            ),
        )

        store = {
            "document": self.schema,
        }

        self.resolver = RefResolver.from_schema(self.schema, store=store)

    # pylint: disable=unused-argument
    def validate(
        self,
        validator: Validator,
        _: Any,
        instance: Any,
        schema: dict[str, Any],
    ) -> ValidationResult:
        # First time child rules are configured against the rule
        # so we can run this now

        if validator.is_type(instance, "string"):
            ssm_validator = validator.evolve(
                context=validator.context.evolve(
                    functions=[],
                    strict_types=True,
                ),
                resolver=self.resolver,
                schema=self.schema,
            )

            instance, errs = decode_str(instance)
            if errs:
                yield ValidationError(
                    "Document is not of type 'object'",
                    validator="type",
                    rule=self,
                )
                return
        else:
            ssm_validator = validator.evolve(
                cfn=validator.cfn,
                context=validator.context.evolve(
                    strict_types=True,
                ),
                resolver=self.resolver,
                schema=self.schema,
            )

        for err in ssm_validator.iter_errors(instance):
            if not err.validator.startswith("fn_") and err.validator not in ["cfnLint"]:
                err.rule = self
            yield err
